class _dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

"""
this allows module wide configuration of drawing methods. The dotdict allows for
convenient access.
Example:

    from swmmio import graphics
    from swmmio.graphics import swmm_graphics as sg

    #configure
    graphics.options.inlcude_parcels = True

    #draws the model with parcels
    sg.drawModel(swmmio.Model(/path/to/model), bbox=su.d68d70)

"""
_o = {
    'include_basemap':False,
    'include_parcels':False,
}

config = _dotdict(_o)



from swmmio.defs.sectionheaders import inp_header_dict, rpt_header_dict
from collections import deque

def random_alphanumeric(n=6):
	import random
	chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
	return ''.join(random.choice(chars) for i in range(n))


def complete_inp_headers (inpfilepath):
    """
    creates a dictionary with all the headers found in an INP file
    (which varies based on what the user has defined in a given model)
    and updates them based on the definitions in inp_header_dict
    this ensures the list is comprehensive

    RETURNS:
        a dictionary including
            'headers'->
                    header section keys and their respective cleaned column headers
            'order' ->
                    an array of section headers found in the INP file
                    that perserves the original order
    """
    foundheaders= {}
    order = []
    #print inp_header_dict
    with open(inpfilepath) as f:
        for line in f:
            if '[' and ']' in line:
                h = line.strip()
                order.append(h)
                if h in inp_header_dict:
                    foundheaders.update({h:inp_header_dict[h]})
                else:
                    foundheaders.update({h:'blob'})

    return {'headers':foundheaders, 'order':order}

def complete_rpt_headers (rptfilepath):
    """
    creates a dictionary with all the headers found in an RPT file
    (which varies based on what the user has defined in a given model)
    and updates them based on the definitions in rpt_header_dict
    this ensures the list is comprehensive

    RETURNS:
        a dictionary including
            'headers'->
                    header section keys and their respective cleaned column headers
            'order' ->
                    an array of section headers found in the RPT file
                    that perserves the original order
    """
    foundheaders= {}
    order = []
    with open(rptfilepath) as f:
        buff3line = deque()
        for line in f:


            #maintains a 3 line buffer and looks for instances where
            #a top and bottom line have '*****' and records the middle line
            #typical of section headers in RPT files
            buff3line.append(line)
            if len(buff3line) > 3:
                buff3line.popleft()

            if ('***********'in buff3line[0] and
				'***********'in buff3line[2] and
				len(buff3line[1].strip()) > 0):
                h = buff3line[1].strip()
                order.append(h)
                if h in rpt_header_dict:
                    foundheaders.update({h:rpt_header_dict[h]})
                else:
                    foundheaders.update({h:'blob'})

    return {'headers':foundheaders, 'order':order}



import pandas as pd
import shutil
import os
import fileinput
import itertools
from datetime import datetime
from swmmio.swmmio import Model
from swmmio.utils import functions as funcs
from swmmio.utils.dataframes import create_dataframeINP
from swmmio.utils import swmm_utils as su
from swmmio.version_control import utils as vc_utils
from swmmio.version_control import inp

#from .utils.text import * #functions for processing inp/rpt/txt files

pd.options.display.max_colwidth = 200


def propagate_changes_from_baseline(baseline_dir, alternatives_dir, combi_dir,
                                    version_id='', comments=''):

    #stuff
    """
    if the baseline model has changes that need to be propogated to all models,
    iterate through each model and rebuild the INPs with the new baseline and
    existing build instructions. update the build instructions to reflect the
    revision date of the baseline.
    """
    version_id += '_' + datetime.now().strftime("%y%m%d%H%M%S")

    #collect the directories of all models
    model_dirs = []
    for alt in os.listdir(alternatives_dir):
        #print alt
        #iterate through each implementation level of each alternative
        for imp_level in os.listdir(os.path.join(alternatives_dir, alt)):
            #create or refresh the build instructions file for the alternatives
            model_dirs.append(os.path.join(alternatives_dir, alt, imp_level))

    model_dirs += [os.path.join(combi_dir, x) for x in os.listdir(combi_dir)]
    #print model_dirs
    baseline = Model(baseline_dir)
    baseinp = baseline.inp.filePath

    for model_dir in model_dirs:
        model = Model(model_dir)
        vc_directory = os.path.join(model_dir, 'vc')
        latest_bi = vc_utils.newest_file(vc_directory)

        #update build instructions metdata and build the new inp
        bi = inp.BuildInstructions(latest_bi)
        bi.metadata['Parent Models']['Baseline'] = {baseinp:vc_utils.modification_date(baseinp)}
        bi.metadata['Log'].update({version_id:comments})
        bi.save(vc_directory, version_id+'.txt')
        print 'rebuilding {} with changes to baseline'.format(model.name)
        bi.build(baseline_dir, model.inp.filePath) #overwrite old inp




def create_combinations(baseline_dir, alternatives_dir, combi_dir, version_id='',
                        comments=''):

    """
    given a set of main alternatives split into models varying levels of
    implementation, this function combines all implementation levels of all
    alternatives into all logical combinations.
    """

    basemodel = Model(baseline_dir)
    baseinp = basemodel.inp.filePath
    alt_directories = os.listdir(alternatives_dir) #list of dirs holding alt models
    implementation_levels = []
    newmodels = []
    version_id += '_' + datetime.now().strftime("%y%m%d%H%M%S")

    for alt in alt_directories:

        #iterate through each implementation level of each alternative
        for imp_level in os.listdir(os.path.join(alternatives_dir, alt)):

            implementation_levels.append(os.path.join(alt, imp_level))

            #create or refresh the build instructions file for the alternatives
            alt_imp_level_dir = os.path.join(alternatives_dir, alt, imp_level)
            alt_imp_inp = Model(alt_imp_level_dir).inp.filePath
            vc_directory = os.path.join(alt_imp_level_dir, 'vc')

            if not os.path.exists(vc_directory):
                print 'creating new build instructions for {}'.format(imp_level)
                inp.create_inp_build_instructions(baseinp, alt_imp_inp, vc_directory,
                                                  version_id, comments)

            else:
                #check if the alternative model was changed since last run of this tool
                #--> compare the modification date to the BI's modification date meta data
                latest_bi = vc_utils.newest_file(vc_directory)
                if not vc_utils.bi_is_current(latest_bi):
                    #revision date of the alt doesn't match the newest build
                    #instructions for this 'imp_level', so we should refresh it
                    print 'updating build instructions for {}'.format(imp_level)
                    inp.create_inp_build_instructions(baseinp, alt_imp_inp,
                                                      vc_directory, version_id,
                                                      comments)


    #creat directories for new model combinations
    for L in range(1, len(implementation_levels)+1):

        #break
        for subset in itertools.combinations(implementation_levels, L):
            #subset e.g. = 'A\A01'

            #newcombi = '_'.join(subset)
            newcombi = '_'.join([os.path.split(s)[1] for s in subset])
            new_combi_dir = os.path.join(combi_dir, newcombi)
            vc_directory = os.path.join(new_combi_dir, 'vc')

            #create a list of the parent directories, use that to prevent
            #two or more from same alternative
            alternative_dirs = [os.path.split(s)[0] for s in subset]

            if len(alternative_dirs) == len(set(alternative_dirs)) and len(subset) > 1:
                #confirming the list length is equal to the set length (hashable)
                #confirms that there are not duplicates in the items list
                parent_vc_dirs = [os.path.join(alternatives_dir, f, 'vc') for f in subset]
                latest_parent_bis = [vc_utils.newest_file(d) for d in parent_vc_dirs]
                build_instrcts = [inp.BuildInstructions(bi) for bi in latest_parent_bis]

                if not os.path.exists(new_combi_dir):#and newcombi not in flavors:
                    #check to make sure new model doesn't repeat two or more from
                    #a particular genre.
                    #print new_combi_dir

                    os.mkdir(new_combi_dir)
                    newinppath = os.path.join(new_combi_dir, newcombi + '.inp')

                    #collect build instructions from each alt's implementation
                    #level for this combination. Select those with the current
                    #version id, or the latest version.
                    print 'creating new child model: {}'.format(newcombi)
                    new_build_instructions = sum(build_instrcts)
                    new_build_instructions.save(vc_directory, version_id+'.txt')
                    new_build_instructions.build(baseline_dir, newinppath)

                else:
                    #check if the alternative model was changed since last run
                    #of this tool --> compare the modification date to the BI's
                    #modification date meta data
                    latest_bi = vc_utils.newest_file(os.path.join(new_combi_dir,'vc'))
                    if not vc_utils.bi_is_current(latest_bi):
                        #revision date of the alt doesn't match the newest build
                        #instructions for this 'imp_level', so we should refresh it
                        print 'updating child build instructions for {}'.format(newcombi)
                        newinppath = os.path.join(new_combi_dir, newcombi + '.inp')
                        new_build_instructions = sum(build_instrcts)
                        new_build_instructions.save(vc_directory, version_id+'.txt')
                        new_build_instructions.build(baseline_dir, newinppath)



from swmmio.swmmio import Model
from swmmio.reporting import reporting
from swmmio.reporting import functions
from swmmio.utils import swmm_utils as su
from time import strftime
import os
import shutil
import math
from itertools import chain

REPORT_DIR_NAME = 'Report'

def batch_cost_estimates(baseline_dir, segments_dir, options_dir, results_file,
                         supplemental_cost_data=None, create_proj_reports=True):
    """
    compute the cost estimate of each model/option in the segments and
    combinations directories. Resulsts will be printed in the results text file.
    """
    #combine the segments and options (combinations) into one iterable
    paths = (segments_dir, options_dir)
    baseline = Model(baseline_dir)

    for path, dirs, files in chain.from_iterable(os.walk(path) for path in paths):

        for f in files:
            if '.inp' in f:
                inp_path = os.path.join(path,f)
                alt = Model(inp_path)

                #calculate the cost
                costsdf = functions.estimate_cost_of_new_conduits(baseline, alt,
                                                                  supplemental_cost_data)
                cost_estimate = costsdf.TotalCostEstimate.sum() / math.pow(10, 6)
                print '{}: ${}M'.format(alt.name, round(cost_estimate,1))

                model_id = os.path.splitext(f)[0]
                with open(results_file, 'a') as res:
                    res.write('{}, {}\n'.format(model_id, cost_estimate))

                if create_proj_reports:
                    #create a option-specific per segment costing csv file
                    report_dir = os.path.join(alt.inp.dir, REPORT_DIR_NAME)
                    fname = '{}_CostEstimate_{}.csv'.format(alt.name, strftime("%y%m%d"))
                    cost_report_path = os.path.join(report_dir, fname)
                    if not os.path.exists(report_dir):os.mkdir(report_dir)
                    costsdf.to_csv(cost_report_path)


def batch_post_process(options_dir, baseline_dir, log_dir, bbox=None, overwrite=False):
    """
    batch process all models in a given directory, where child directories
    each model (with .inp and .rpt companions). A bbox should be passed to
    control where the grahics are focused. Specify whether reporting content
    should be overwritten if found.
    """
    baseline = Model(baseline_dir)
    folders = os.listdir(options_dir)
    logfile = os.path.join(log_dir, 'logfile.txt')
    with open (logfile, 'a') as f:
        f.write('MODEL,NEW_SEWER_MILES,IMPROVED,ELIMINATED,WORSE,NEW\n')
    for folder in folders:
        #first check if there is already a Report directory and skip if required
        current_dir = os.path.join(options_dir, folder)
        report_dir = os.path.join(current_dir, REPORT_DIR_NAME)
        if not overwrite and os.path.exists(report_dir):
            print 'skipping {}'.format(folder)
            continue

        else:
            #generate the report
            current_model = Model(current_dir)
            print 'Generating report for {}'.format(current_model.inp.name)
            #reporting.generate_figures(baseline, current_model, bbox=bbox, imgDir=report_dir, verbose=True)
            report = reporting.Report(baseline, current_model)
            report.write(report_dir)

            #keep a summay log
            with open (logfile, 'a') as f:
                #'MODEL,NEW_SEWER_MILES,IMPROVED,ELIMINATED,WORSE,NEW'
                f.write('{},{},{},{},{},{}\n'.format(
                                                    current_model.inp.name,
                                                    report.sewer_miles_new,
                                                    report.parcels_flooding_improved,
                                                    report.parcels_eliminated_flooding,
                                                    report.parcels_worse_flooding,
                                                    report.parcels_new_flooding
                                                    )
                                                )

def gather_files_in_dirs(rootdir, targetdir, searchfilename, newfilesuffix='_Impact.png'):

    """
    scan through a directory and copy files having a given file name into a
    taget directory. This is useful when wanting to collect the same report image
    within a SWMM model's Report directory (when there are many models).

    This expects the Parent<Parent directory of a file to be called the unique
    model ID. For example, the png in the following dir:

        rootdir > ... > M01_R01_W02 > Report > '03 Impact of Option.png'

    would be copied as follows:

        targetdir > 'M01_R01_W02_Impact.png'

    """


    for root, dirs, files in os.walk(rootdir):

        for f in files:
            #if the file name = the searched file name, copy it to the target dirs
            #for example searchfilename = '03 Impact of Option.png'
            if os.path.basename(f) == searchfilename:
                current_dir = os.path.dirname(os.path.join(root, f))
                model_id = os.path.basename(os.path.dirname(current_dir))

                newf = os.path.join(targetdir, model_id + newfilesuffix)
                shutil.copyfile(os.path.join(root, f), newf)



from swmmio.run_models import run
from swmmio.swmmio import Model
from swmmio.reporting import reporting
from swmmio.utils import swmm_utils as su
from swmmio.reporting import batch
from multiprocessing import Pool, cpu_count
from datetime import datetime
import os
import sys
wd = os.getcwd()

def run_swmm_engine(inp_folder):
    now = datetime.now().strftime("%y%m%d_%H%M")
    logfile = os.path.join(wd, 'log_'+now+'.txt')

    m = Model(inp_folder)
    print 'inp_folder = {}'.format(inp_folder)
    if not m.rpt:
        # print 'completed {} at {}'.format(m.inp.name, datetime.now())
        with open (logfile, 'a') as f:
            now = datetime.now().strftime("%y-%m-%d %H:%M")
            f.write('{} -- {} started... '.format(now, m.inp.name))
            run.run_hot_start_sequence(m)
            now = datetime.now().strftime("%y-%m-%d %H:%M")
            f.write('completed at {}\n'.format(now))
    else:
        with open (logfile, 'a') as f:
            f.write('RPT already exist: {}\n'.format(m.rpt.filePath))

def main(dirs_containing_inps, cores_left):

    """
    called from the cli:
    swmmio -sp DIR1, DIR2, ... -cores_left=4
    """

    # create multiprocessing Pool object using all cores less the -cores_left
    #beware of setting -cores_left=0, CPU usage will max the F out
    pool = Pool(cpu_count() - cores_left)

    #create a process pool with the run_swmm_engine() func on each directory
    res = pool.map(run_swmm_engine, dirs_containing_inps)


if __name__ == '__main__':
    """
    the old way
    """
    for arg in sys.argv[1:]:
        print arg
    args = sys.argv[1:]
    if len(args) > 0:
        combi_folder = args[0]
        folders = os.listdir(combi_folder)
        dirs_containing_inps = []
        for root, dirs, files in os.walk(combi_folder):
            for file in files:
                if file.endswith('.inp') and 'bk' not in root:
                    #we've found a directory containing an inp
                    print file
                    dirs_containing_inps.append(root)
                    #dirs_containing_inps.append(os.path.join(root, file))

        print 'dirs_containing_inps = {}'.format(dirs_containing_inps)

        pool = Pool(cpu_count() - 6) # use all but 6 available cores
        # proces data_inputs iterable with pool
        res = pool.map(run_swmm_engine, dirs_containing_inps)

        print "hi im done running"
        baseline_dir = r'F:\models\SPhila\MasterModels_170104\Baseline' #r'P:\06_Tools\v_control\BIG\Baseline'
        log_dir = r'F:\models\SPhila\MasterModels_170104\ProjectAdmin'
        # batch.batch_post_process(combi_folder, baseline_dir, log_dir, bbox=su.d68d70)



from datetime import datetime
import os
import json
import pandas as pd
from swmmio.utils import functions


def write_inp_section(file_object, allheaders, sectionheader, section_data, pad_top=True, na_fill = ''):

    """
    given an open file object, excelwriter object, list of header sections, the curent
    section header, and the section data in a Pandas Dataframe format, this function writes
    the data to the file object.
    """

    f = file_object
    add_str =  ''
    if not section_data.empty:
        if pad_top:
            f.write('\n\n' + sectionheader + '\n') #add SWMM-friendly header e.g. [DWF]
        else:
            f.write(sectionheader + '\n')
        if allheaders and allheaders['headers'].get(sectionheader, 'blob') == 'blob':
            #to left justify based on the longest string in the blob column
            formatter = '{{:<{}s}}'.format(section_data[sectionheader].str.len().max()).format
            add_str = section_data.fillna('').to_string(
                                                        index_names=False,
                                                        header=False,
                                                        index=False,
                                                        justify='left',
                                                        formatters={sectionheader:formatter}
                                                        )
            # #write section to excel sheet
            # sheetname = sectionheader.replace('[', "").replace(']', "")
            # section_data.to_excel(excelwriter, sheetname, index=False)

        else:
            #naming the columns to the index name so the it prints in-line with col headers
            f.write(';;')
            #to left justify on longest string in the Comment column
            #this is overly annoying, to deal with 'Objects' vs numbers to remove
            #two bytes added from the double semicolon header thing (to keep things lined up)
            objectformatter =   {hedr:' {{:<{}}}'.format(section_data[hedr].apply(str).str.len().max()).format
                                    for hedr in section_data.columns}
            numformatter =      {hedr:'  {{:<{}}}'.format(section_data[hedr].apply(str).str.len().max()).format
                                    for hedr in section_data.columns if section_data[hedr].dtype!="O"}
            objectformatter.update(numformatter)
            add_str = section_data.fillna(na_fill).to_string(
                                                        index_names=False,
                                                        header=True,
                                                        justify='left',
                                                        formatters=objectformatter#{'Comment':formatter}
                                                        )

            # #write section to excel sheet
            # sheetname = sectionheader.replace('[', "").replace(']', "")
            # section_data.to_excel(excelwriter, sheetname)


        #write the dataframe as a string
        f.write(add_str + '\n\n')
def write_meta_data(file_object, metadicts):

    file_object.write(json.dumps(metadicts, indent=4))
    file_object.write('\n' + '='*100 + '\n\n')

def read_meta_data(filepath):

    s = ''
    with open(filepath) as file_object:
        for line in file_object:
            if '================' in line:
                break
            s += line.strip()

    return json.loads(s)

def bi_is_current(build_instr_file):
    """
    check if a given build instruction file has any parent models whose
    date modified does not match the date modified of the parent INP file
    """

    meta = read_meta_data(build_instr_file)
    baseline = meta['Parent Models']['Baseline']
    alternatives = meta['Parent Models']['Alternatives']
    #parents = baseline.update(alternatives)
    # print meta['Parent Models']['Baseline']
    # print alternatives
    for inp, revisiondate in baseline.iteritems():
        if modification_date(inp) != revisiondate:
            return False

    for inp, revisiondate in alternatives.iteritems():
        if modification_date(inp) != revisiondate:
            return False

    return True

def bi_latest_parent_date_modified(vc_dir, parentname):
    """
    given a path to a version control directory of build instructions and the
    name of the parent model, return the parent model's revision date
    """
    newest_bi = newest_file(vc_dir)
    meta = read_meta_data(newest_bi)
    # with open (newest_bi) as f:


    return meta['Parent Models'][parentname]


def newest_file(directory):
    """
    return the newest file (most recent) in a given directory. Beware that
    people report the min / max to do different things per OS...
    """
    files = os.listdir(directory)
    return max([os.path.join(directory, f) for f in files], key=os.path.getctime)

def modification_date(filename, string=True):
    """
    get modification datetime of a file
    credit: Christian Oudard
    'stackoverflow.com/questions/237079/how-to-get-file-creation-modification-
    date-times-in-python'
    """
    t = os.path.getmtime(filename)
    dt = datetime.fromtimestamp(t)
    if string:
        return dt.strftime("%y-%m-%d %H:%M")
    else:
        return dt#datetime.fromtimestamp(t)

def write_excel_inp_section(excelwriter, allheaders, sectionheader, section_data):

    if not section_data.empty:
        sheetname = sectionheader.replace('[', "").replace(']', "")

        if allheaders['headers'].get(sectionheader, 'blob') == 'blob' and not section_data.empty:
            #write section to excel sheet
            section_data.to_excel(excelwriter, sheetname, index=False)
        else:
            #clean up the format a bit
            # workbook = excelwriter.book
            # elem_fmt = workbook.add_format({'align': 'left'})
            # worksheet = excelwriter.sheets[sheetname]
            # worksheet.set_column('A:A', 20)

            section_data.to_excel(excelwriter, sheetname)

def create_change_info_sheet(excelwriter, model1, model2, name='', message=''):

    ix =  ['Date', 'BaselineModel', 'TargetModel', 'Commit', 'Name', 'Comment']
    vals =  [
            datetime.now().strftime("%y-%m-%d %H:%M"),
            model1.inp.filePath,
            model2.inp.filePath,
            functions.random_alphanumeric(12),
            name,
            message
            ]
    df = pd.DataFrame(data=vals, index=ix,columns=['FileInfo'])
    df.to_excel(excelwriter, 'FileInfo')


def create_info_sheet(excelwriter, basemodel, parent_models=[]):

    #create an info sheet for the Excel file
    timeofcreation = datetime.now()
    s = pd.Series([datetime.now(), basemodel.inp.filePath] + [x.inp.filePath for x in parent_models] )
    s.index = ['DateCreated', 'Basemodel']+['ParentModel_' + str(i) for i,x in enumerate(parent_models)]
    df = pd.DataFrame(s, columns=['FileInfo'])
    df.to_excel(excelwriter, 'FileInfo')




#Module is intended to provide high level access to post processing routines
#such that standard reporting and figures can be generated to report on the
#perfomance of given SFR alternatives/options

from swmmio.reporting import compare_models as scomp
from swmmio.reporting import functions
from swmmio.utils import swmm_utils as su
from swmmio.graphics import swmm_graphics as sg
from swmmio.graphics import draw_utils as du
from swmmio.utils.dataframes import create_dataframeRPT
from swmmio import parcels as p
import os
import math
import pandas as pd
from definitions import *



class Report(object):

    def __init__(self, baseline_model, option, existing_parcel_flooding = None,
                 additional_costs=None):

        #parcel_features = 'PWD_PARCELS_SHEDS_PPORT' #'PWD_PARCELS_SHEDS'
        #parcel calculations
        anno_results={}
        #if not existing_parcel_flooding:
        existing_parcel_flooding = p.parcel_flood_duration(baseline_model, parcel_features=PARCEL_FEATURES,
    											bbox=None, anno_results=anno_results)['parcels']

        proposed_parcel_flooding = p.parcel_flood_duration(option, parcel_features=PARCEL_FEATURES,
												bbox=None, anno_results=anno_results)['parcels']

        delta_parcels = p.compareParcels(existing_parcel_flooding, proposed_parcel_flooding,
										bbox=None, floodthreshold=0.0833,
										delta_threshold=0.25,
										anno_results=anno_results)

        self.baseline=baseline_model
        self.option = option
        self.delta_parcels = delta_parcels['parcels']
        self.anno_results = anno_results
        self.sewer_miles_new = functions.length_of_new_and_replaced_conduit(baseline_model, option)  / 5280.0
        self.sewer_miles_replaced = None
        self.storage_new = None #some volume measurement
        self.cost_estimate_data = functions.estimate_cost_of_new_conduits(baseline_model, option, additional_costs)
        self.cost_estimate = self.cost_estimate_data.TotalCostEstimate.sum() / math.pow(10, 6)
        self.parcels_new_flooding = delta_parcels['new_flooding']
        self.parcels_worse_flooding = delta_parcels['increased_flooding']
        self.parcels_eliminated_flooding = delta_parcels['eliminated_flooding']
        self.parcels_flooding_improved = delta_parcels['decreased_flooding'] + delta_parcels['eliminated_flooding']

        print '{} new miles '.format(self.sewer_miles_new)
        print '{} parcels deltas'.format(len(self.delta_parcels))

    def write (self, report_dir):
        #generate the figures
        generate_figures(self.baseline, self.option,
                        delta_parcels = self.delta_parcels,
                        anno_results = self.anno_results,
                        bbox = su.d68d70, imgDir=report_dir)

        #write the cost estimate file
        costsdf  = self.cost_estimate_data[['Length', 'Shape', 'Geom1', 'Geom2',
                                            'Geom3', 'XArea', 'Volume',
                                            'CostEstimate', 'AdditionalCost',
                                            'AddCostNotes', 'TotalCostEstimate']]
        costsdf.to_csv(os.path.join(report_dir, self.option.inp.name + '_cost_data.csv'))
        #write text report
        with open(os.path.join(report_dir, 'report.txt'), 'w') as f:
            f.write('Parcels Improved {}\n'.format(self.parcels_flooding_improved))
            f.write('New Sewer Miles {}\n'.format(self.sewer_miles_new))
            f.write('Cost Estimate {}\n'.format(self.cost_estimate))

def generate_figures(model1, model2, delta_parcels=None, anno_results = {},
                    bbox=None, imgDir=None, verbose=False):

    #calculate the bounding box of the alternative/option
    #such that figures can be zoomed to the area
    #bbox = scomp.extents_of_changes(model1, model2, extent_buffer=1.0)

    #SIMPLE PLAN VIEW OF OPTION (showing new conduits)
    imgname = "00 Proposed Infrastructure"
    scomp.drawModelComparison(model1, model2, nodeSymb=None, parcelSymb=None,
                                bbox=bbox, imgName=imgname, imgDir=imgDir)

    #EXISTING CONDITIONS PARCEL FLOOD DURATION
    imgname = "01 Existing Parcel Flood Duration"
    if not imgDir:
        imgDir=os.path.join(model2.inp.dir, 'img')
    sg.drawModel(model1, conduitSymb=None, nodeSymb=None, bbox=bbox,
                 imgName=imgname, imgDir=imgDir)

    #PROPOSED CONDITIONS FLOOD DURATION
    imgname = "02 Proposed Parcel Flood Duration"
    sg.drawModel(model2, conduitSymb=None, nodeSymb=None, bbox=bbox,
                  imgName=imgname, imgDir=imgDir)

    #IMPACT OF INFRASTRUCTURE
    imgname = "03 Impact of Option"
    scomp.drawModelComparison(model1, model2,
                                delta_parcels = delta_parcels,
                                anno_results = anno_results,
                                nodeSymb=None, bbox=bbox,
                                imgName=imgname, imgDir=imgDir)

    #IMPACT OF INFRASTRUCTURE STUDY-AREA-WIDE
    imgname = "04 Impact of Option - Overall"
    scomp.drawModelComparison(model1, model2,
                                delta_parcels = delta_parcels,
                                anno_results = anno_results,
                                nodeSymb=None, bbox=su.study_area,
                                imgName=imgname, imgDir=imgDir)

    #IMPACT OF INFRASTRUCTURE CHANGE OF PEAK FLOW
    imgname = "05 Hydraulic Deltas"
    scomp.drawModelComparison(model1, model2,
                                delta_parcels = delta_parcels,
                                anno_results = anno_results,
                                conduitSymb=du.conduit_options('compare_flow'),
                                bbox=bbox, imgName=imgname, imgDir=imgDir)


def timeseries_join(elements, *models):
    """
    Given a list of element IDs and Model objects, a dataframe is returned
    with the elements' time series data for each model.
    Example:
        df = timeseries_join(elements = ['P1, P2'], model1, model2)
        returns df with index = DateTime and cols = [model1_P1_FlowCFS, model1_P2_FlowCFS,
                                                    model2_P1_FlowCFS, model2_P2_FlowCFS]
    """
    # dfs = [[dataframes.create_dataframeRPT(m.rpt, 'Link Results', el)[['FlowCFS']]
    #         for el in elements] for m in models]
    param_name = 'FlowCFS' #as named by our dataframe method
    section_name = 'Link Results' #in the rpt
    dfs = [create_dataframeRPT(m.rpt, section_name, elem)[[param_name]].rename(
                columns={param_name:'{}_{}_{}'.format(m.inp.name, elem, param_name)}
                )
                for m in models for elem in elements]

    df = pd.concat(dfs, axis=1)
    return df



#module to contain general function pertaining to graphics, (colors, PIL stuff, constants, and options)
#drawing options
import math
from swmmio.graphics import config
from definitions import PARCEL_FEATURES, GEODATABASE
#COLOR DEFS
red = 		(250, 5, 5)
blue = 		(5, 5, 250)
lightblue = (184, 217, 242)
shed_blue = (0,169,230)
white =		(250,250,240)
black = 	(0,3,18)
mediumgrey = (190, 190, 180)
lightgrey = (235, 235, 225)
grey = 		(100,95,97)
park_green = (115, 178, 115)
green = 	(115, 220, 115)
green_bright= (23, 191, 23)
lightgreen = (150,212,166)
water_grey = (130, 130, 130)
purple = (250, 0, 250)


def basemap_options(**kwargs):
	basemap_options = {
    'gdb': GEODATABASE,
    'features': [
		#this is an array so we can control the order of basemap layers
        # {
            # 'feature': 'D68_Dissolve',
            # 'fill': None,
            # 'outline': su.shed_blue,
            # 'featureDict': None,
            # 'cols': ["OBJECTID", "SHAPE@"]
		# },
        # {
            # 'feature': 'PWD_PARCELS_SPhila_ModelSpace',
            # 'fill': su.lightgrey,
            # 'outline': None,
            # 'featureDict': None,
            # 'cols': ["OBJECTID", "SHAPE@"]
		# },
		# {
            # 'feature': 'SouthPhilaModelExtents',
            # 'fill': su.shed_blue,
            # 'outline': None,
            # 'featureDict': None,
            # 'cols': ["OBJECTID", "SHAPE@"]
		# },
		# {
            # 'feature': 'detailedsheds',
            # 'fill': None,
            # 'outline': su.red,
            # 'featureDict': None,
            # 'cols': ["OBJECTID", "SHAPE@"]
		# },

		{
            'feature': 'PhiladelphiaParks',
            'fill': park_green,
            'outline': None,
            'featureDict': None,
            'cols': ["OBJECTID", "SHAPE@"]
		},
        {
            'feature': 'HydroPolyTrim',
            'fill':water_grey,
            'outline': None,
            'featureDict': None,
            'cols': ["OBJECTID", "SHAPE@"]
		},
        {
            'feature': 'Streets_Dissolved5_SPhilly',
            'fill': lightgrey,
            'width': 0,
            'fill_anno': grey,
            'outline': None,
            'featureDict': None,
            'cols': ["OBJECTID", "SHAPE@", "ST_NAME"]
		}
      ],
	}
	feats = []
	for key, value in kwargs.iteritems():
		basemap_options.update({key:value})

	return basemap_options
def conduit_options(type, **kwargs):
	#drawing options for conduits
	conduit_def_symbologies = {
		'stress': {
			'title': 'Condiut Stress',
			'description': 'Shows how taxed conduits are based on their flow (peak flow) with respect to their full-flow capacity',
			'threshold': 1,#fraction used
			'type': 'stress',
			'fill':greyRedGradient,
			'draw_size':line_size,
			'exp':0.8,
			'xplier':10
		},
		'compare_flow': {
			'title': 'Flow Comparison',
			'description': 'Shows the change in peak flows in conduits between a baseline and proposed model',
			'type': 'compare_flow',
			'fill':greyRedGradient,
			'draw_size':line_size,
			'exp':0.67,
			'xplier':1
		},
		'compare_hgl': {
			'title': 'HGL Comparison',
			'description': 'Shows the change in HGL in conduits between a baseline and proposed model',
			'type': 'compare_hgl',
			'fill':greyRedGradient,
			'draw_size':line_size,
			'exp':1,
			'xplier':10
		},
		'proposed_simple': {
			'title': 'Proposed Infrastructure',
			'description': 'Shows the new infrastructure only',
			'type': 'proposed_simple',
			'fill':blue
		},
		'capacity_remaining': {
			'title': 'Remaining Capacity',
			'description': 'Shows the amount of full flow capacity remaining in conduits',
			'type': 'capacity_remaining',
			'fill':blue,
			'draw_size':line_size,
			'exp':0.75,
			'xplier':1
		},
		'flow': {
			'title': 'Condiut Flow',
			'description': 'Shows the flow in conduits with proportional ine weight',
			'type': 'flow',
			'fill':blue,
			'draw_size':line_size,
			'exp':0.67,
			'xplier':1
		},
		'flow_stress': {
			'title': 'Condiut Flow & Stress',
			'description': 'Shows the flow in conduits with line weight and color based on "stress" (flow/full-flow capacity)',
			'type': 'flow_stress',
			'fill':greenRedGradient,
			'draw_size':line_size,
			'exp':0.67,
			'xplier':1
		},
		'trace': {
			'title': 'Trace Upstream',
			'description': '',
			'type': 'trace'
		}
	}

	selected_ops = conduit_def_symbologies[type]
	for key, value in kwargs.iteritems():
		selected_ops.update({key:value})

	return selected_ops
def node_options(type='flood', **kwargs):

	#drawing options for conduits
	node_symbologies = {
		'flood': {
			'title': 'Node Flood Duration',
			'description': 'Shows the node duration proporationally in size',
			'threshold': 0.083,#minutes,
			'fill': red,
			'type': 'flood'
		},
		'flood_color': {
			'title': 'Node Flood Duration',
			'description': 'Shows the node duration via color gradient in size',
			'threshold': 0.083,#minutes,
			'fill':greenRedGradient,
			'type': 'flood_color'
		}
	}

	selected_ops = node_symbologies[type]
	for key, value in kwargs.iteritems():
		selected_ops.update({key:value})

	return selected_ops
def parcel_options(type='flood', **kwargs):

	#drawing options for conduits
	parcel_symbologies = {
		'flood': {
				'title': 'Parcel Flood Duration',
				'description': 'Shows the parcels flood duration severity based on color',
				'threshold': 0.08333,
				'delta_threshold':0.25,
				'fill': red,
				'outline': None,
				'type': 'flood',
				'feature':PARCEL_FEATURES,
				'gdb':GEODATABASE
				},
		'compare_flood': {
				'title': 'Parcel Flood Change',
				'description': 'Shows the parcels flood duration severity based on color',
				'threshold': 0.08333,
				'delta_threshold':0.25,
				'fill': red,
				'outline': None,
				'type': 'compare_flood',
				'feature':PARCEL_FEATURES,
				'gdb':GEODATABASE
				}
			}

	selected_ops = parcel_symbologies[type]
	for key, value in kwargs.iteritems():
		selected_ops.update({key:value})

	return selected_ops
def default_draw_options():

	if config.include_basemap:
		basemap_symbology = basemap_options()
	else:
		basemap_symbology = None

	if config.include_parcels:
		parcels_symbology = parcel_options('flood')
	else:
		parcels_symbology = None

	default_options = {
		'width': 2048,
		'bbox':None,
		'imgName':None,
		'imgDir':None,
		'nodeSymb': node_options('flood'),
		'conduitSymb': conduit_options('stress'),
		'basemap': basemap_symbology,
		'parcelSymb': parcels_symbology,
		'bg': white,
		'xplier': 1,
		'traceUpNodes': [],
		'traceDnNodes': [],
		'fps': 7.5,
		'title': None
	}
	return default_options


#DRAWING UTILITY FUNCTIONS
def greenRedGradient(x, xmin, xmax):

	range = xmax - xmin
	scale = 255 / range

	x = min(x, xmax) #limit any vals to the prescribed max

	#print "range = " + str(range)
	#print "scale = " + str(scale)
	r = int(round(x*scale))
	g = int(round(255 - x*scale))
	b = 0

	return (r, g, b)
def greyRedGradient(x, xmin, xmax):

	range = xmax - xmin

	rMin = 100
	bgMax = 100
	rScale = (255 - rMin) / range
	bgScale = (bgMax) / range
	x = min(x, xmax) #limit any vals to the prescribed max


	#print "range = " + str(range)
	#print "scale = " + str(scale)
	r = int(round(x*rScale + rMin ))
	g = int(round(bgMax - x*bgScale))
	b = int(round(bgMax - x*bgScale))

	return (r, g, b)
def greyGreenGradient(x, xmin, xmax):

	range = xmax - xmin

	gMin = 100
	rbMax = 100
	gScale = (255 - gMin) / range
	rbScale = (rbMax) / range
	x = min(x, xmax) #limit any vals to the prescribed max


	#print "range = " + str(range)
	#print "scale = " + str(scale)
	r = int(round(rbMax - x*rbScale))
	g = int(round(x*rbScale + gMin ))
	b = int(round(rbMax - x*rbScale))

	return (r, g, b)

def col2RedGradient(x, xmin, xmax, startCol=lightgrey):

	range = xmax - xmin

	rMin = startCol[0]
	gMax = startCol[1]
	bMax = startCol[2]

	rScale = (255 - rMin) / range
	gScale = (gMax) / range
	bScale = (bMax) / range
	x = min(x, xmax) #limit any vals to the prescribed max


	#print "range = " + str(range)
	#print "scale = " + str(scale)
	r = int(round(x*rScale + rMin ))
	g = int(round(gMax - x*gScale))
	b = int(round(bMax - x*bScale))

	return (r, g, b)

	#lightgrey = (235, 235, 225)

def blueRedGradient(x, xmin, xmax):

	range = xmax - xmin
	scale = 255 / range

	x = min(x, xmax) #limit any vals to the prescribed max

	#print "range = " + str(range)
	#print "scale = " + str(scale)
	r = int(round(x*scale))
	g = 0
	b = int(round(255 - x*scale))

	return (r, g, b)
def line_size(q, exp=1):
	return int(round(math.pow(q, exp)))

#geometry related functions
def circleBBox(coordinates, radius):
	#returns the bounding box of a circle given as centriod coordinate and radius
	x = coordinates[0] #this indexing is because other elements haev more than on coordinate (ulgy pls fix)
	y = coordinates[1]
	r = radius

	return (x-r, y-r, x+r, y+r)

def getX2(y1, y2, length, x1=0):

	#return the x2 coordinate given y1, y2, the line segment length, and x0

	a = y2 - y1
	c = length
	return math.sqrt(c*c - a*a) + x1



#test modules
import unittest
import sys
sys.path.append(r'P:\06_Tools\swmmio')
from swmmio import swmm_headers

#check that header match and replace somethig in the file
class sectionheadersTest(unittest.TestCase):
    def test(self):
        with open(r'sample_models\sample.inp') as f:
            #print f.read(1000)
            text = f.read()
            for hpair in swmm_headers.inpHeaderList:
                print hpair[0][:200]
                matched_an_inp_header = False
                #hpair[0]
                if hpair[0] in text:
                    matched_an_inp_header = True

        self.assertTrue(matched_an_inp_header)

class RptHeadersTest(unittest.TestCase):
    def test(self):
        with open(r'sample_models\sample.rpt') as f:
            #print f.read(1000)
            text = f.read()
            for hpair in swmm_headers.rptHeaderList:
                #print hpair[0][20:]
                matched_an_rpt_header = False
                #hpair[0]
                if hpair[0] in text:
                    matched_an_rpt_header = True

        self.assertTrue(matched_an_rpt_header)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#coding:utf-8

#graphical functions for SWMM files
#import swmmio
from swmmio.utils import swmm_utils as su
from swmmio.graphics import draw_utils as du
from swmmio.graphics import config
from swmmio import parcels
import re
import os
import numpy
from PIL import Image, ImageDraw, ImageFont
import datetime
from datetime import timedelta
from time import gmtime, strftime
import pickle



def saveImage(img, model, imgName=None, imgDir=None, antialias=True, open=False, fileExt=".png", verbose=False):

	#get the size from the Image object
	imgSize = (img.getbbox()[2], img.getbbox()[3])

	#if imgName not specified, define as model name
	if not imgName:
		imgName = model.inp.name

	#create the saving location as necessary
	if not imgDir:
		inp = model.inp
		rpt = model.rpt
		standardDir = os.path.join(inp.dir, "img")
		if not os.path.exists(standardDir):
			#if no directory is specified and none exists at
			#standard location create a new directory
			os.makedirs(os.path.join(inp.dir, "img"))
		newFile = os.path.join(standardDir, imgName) + fileExt
	else:
		#imDir is specified by user
		if not os.path.exists(imgDir):
			#if directory doesn't exist, create new
			os.makedirs(imgDir)
		newFile = os.path.join(imgDir, imgName) + fileExt

	if verbose: print "saving image to: " + newFile

	#shrink for antialiasing niceness (though this blows up the file about 5x)
	if antialias:
		size = (int(imgSize[0]*0.5), int(imgSize[1]*0.5))
		img.thumbnail(size, Image.ANTIALIAS)

	img.save(newFile)
	if open:
		os.startfile(newFile)


def createFeaturesDict(options={},  bbox=None, shiftRatio=None, width=1024):

	#create dictionary containing draw coordinate data for the given basemap options

	gdb = options['gdb']
	features = options['features']

	import arcpy
	if arcpy.Exists(gdb):
		#if the gdb exists, then we can move forward
		#for feature, data in features.iteritems():
		for featureOps in features:
			feature = featureOps['feature']
			if arcpy.Exists(os.path.join(gdb, feature)):

				#this check so that we don't repeat this heavy op over an over
				featureDict = su.shape2Pixels(feature, bbox=bbox, targetImgW=width, where=None, cols=featureOps['cols'],
														shiftRatio=shiftRatio, gdb=gdb)

				featureOps.update({'featureDict':featureDict}) #retain this for later if necessary
				#featureOps['featureDict'].update{featureDict}) #retain this for later if necessary
			else:
				print '{} not found'.format(feature)
	else:
		print '{} not found'.format(gdb)
		return None

	return features


def drawParcels(draw, parcel_flooding_results,  options={}, bbox=None, width=1024, shiftRatio=None):

	parcels_pixels = su.shape2Pixels("PWD_PARCELS_SPhila_ModelSpace", where=None,
									cols = ["PARCELID", "SHAPE@"], targetImgW=width,
									shiftRatio=shiftRatio, bbox=bbox)

	#if 'parcels' in parcel_flooding_results:
	#	parcel_flooding_results = parcel_flooding_results['parcels'] #HACKKK
	threshold = options['threshold']
	newflood = moreflood = floodEliminated= floodlowered= 0
	for PARCELID, parcel in parcel_flooding_results.iteritems():
		fill = du.lightgrey #default
		try:
			if parcel.is_delta:
				#we're dealing with "compare" dictionary
				if parcel.delta_type == 'increased_flooding':
					#parcel previously flooded, now floods more
					fill = du.red
					moreflood += 1

				if parcel.delta_type == 'new_flooding':
					#parcel previously did not flood, now floods in proposed conditions
					fill = du.purple
					newflood += 1

				if parcel.delta_type == 'decreased_flooding':
					#parcel flooding problem decreased
					fill =du.lightblue #du.lightgrey
					floodlowered += 1

				if parcel.delta_type == 'eliminated_flooding':
					#parcel flooding problem eliminated
					fill =du.lightgreen
					floodEliminated += 1

			elif parcel.flood_duration > threshold:
				fill = du.col2RedGradient(parcel.flood_duration + 0.5, 0, 3)

			parcel_pix = parcels_pixels['geometryDicts'][PARCELID]
			draw.polygon(parcel_pix['draw_coordinates'], fill=fill, outline=options['outline'])
		except:
			pass

	#print "newflood = {}\nmoreflood={}\nfloodEliminated={}\nfloodlowered={}".format(newflood, moreflood, floodEliminated, floodlowered)
	#saveImage(img, None, imgName=model.inp.name + "_parcels", imgDir=r'C:\Users\Adam.Erispaha\Desktop\S Phila SWMM\img')

def drawBasemap(draw, img=None, featureDicts=None, options={}, bbox=None, shiftRatio=None, width=1024, xplier=1):


	if not featureDicts:
		#generate the dict containing drawable data
		featureDicts = createFeaturesDict(options,  bbox=bbox, shiftRatio=shiftRatio, width=width) #will return None if probs finding data

	#if featureDicts:
	#	features = featureDicts['features']

	polyDrawCount = 0
	anno_streets = []
	#for feature, data in features.iteritems():
	for feature in featureDicts:
		featureDict = feature['featureDict']
		for poly, polyData in featureDict['geometryDicts'].iteritems():
			if polyData['geomType'] == 'polygon':
				draw.polygon(polyData['draw_coordinates'], fill=feature['fill'], outline=feature['outline'])
				polyDrawCount += 1

			elif polyData['geomType'] == 'polyline':

				draw.line(polyData['draw_coordinates'], fill=feature['fill'], width=feature['width']*xplier)
				if 'ST_NAME' in polyData:
					su.annotateLine(img, polyData, annoKey='ST_NAME', labeled = anno_streets)
					polyDrawCount += 1

def drawModel (model, **kwargs):

	#unpack and update the options
	ops = du.default_draw_options()
	for key, value in kwargs.iteritems():
		ops.update({key:value})
	#return ops
	width = ops['width']
	xplier = ops['xplier']
	bbox = ops['bbox']
	imgName = ops['imgName'] # for some reason saveImage() won't take the dict reference
	imgDir = ops['imgDir'] # for some reason saveImage() won't take the dict reference

	focusConduits = []
	for node in ops['traceUpNodes']:
		#return list of elements upstream of node
		focusConduits += su.traceFromNode(model, node, mode='up')['conduits']

	for node in ops['traceDnNodes']:
		#return list of elements downstream of node
		focusConduits += su.traceFromNode(model, node, mode='down')['conduits']

	#antialias X2
	xplier *= width/1024 #scale the symbology sizes
	width = width*2

	#parse out the main objects of this model
	inp = model.inp
	rpt = model.rpt

	#organize relavant data from SWMM files, #FIX TO NOT RUN ALL THIS IF NOT NECESSARY
	conduitData = model.organizeConduitData(bbox) #dictionary of overall model data, dimension, and the conduit dicts
	conduits = conduitData['conduit_objects']
	pixelData = su.convertCoordinatesToPixels(conduits, targetImgW=width, bbox=bbox)
	shiftRatio = pixelData['shiftRatio']
	imgSize = pixelData['imgSize']

	if not bbox:
		bbox = pixelData['boundingBox'] #this is the box tightly wrapping all conduits, will be <= the user provided bbox

	#node data
	nodeData = model.organizeNodeData(bbox)
	nodes = nodeData['node_objects']
	su.convertCoordinatesToPixels(nodes, targetImgW=width, bbox=bbox)

	img = Image.new('RGB', imgSize, ops['bg'])
	draw = ImageDraw.Draw(img)

	anno_results = {}

	#DRAW THE PARCELS
	if ops['parcelSymb']:

		#grab the parcel&node dict that associates node IDs to each parcel
		parcel_flooding_results = parcels.parcel_flood_duration(model, parcel_features=ops['parcelSymb']['feature'],
															threshold=ops['parcelSymb']['threshold'],
															gdb= ops['parcelSymb']['gdb'], bbox=None,
															anno_results=anno_results)

		drawParcels(draw, parcel_flooding_results['parcels'], options=ops['parcelSymb'], bbox=bbox, width=width, shiftRatio=shiftRatio)

	#DRAW THE BASEMAP
	if ops['basemap']:
		drawBasemap(draw, img=img, options=ops['basemap'], width=width, bbox=bbox, shiftRatio=shiftRatio, xplier = xplier)

	drawCount = 0

	#DRAW THE CONDUITS
	if ops['conduitSymb']:
		for id, conduit in conduits.iteritems():

			if conduit.coordinates: #has coordinate? draw. protect from rdii junk
				su.drawConduit(conduit, draw, ops['conduitSymb'], rpt=rpt, xplier = xplier, highlighted=focusConduits)
				drawCount += 1

	#DRAW THE NODES
	if ops['nodeSymb']:
		for id, node in nodes.iteritems():
			if node.coordinates: #this prevents draws if no flow is supplied (RDII and such)
				#drawNode(node, draw, options, rpt=None, dTime=None, xplier=1):
				su.drawNode(node, draw, rpt=rpt, options=ops['nodeSymb'], dTime=None, xplier=xplier)
				drawCount += 1


	#if conduitDrawCount > 0 and conduitDrawCount % 2000 == 0: print str(conduitDrawCount) + " pipes processed "
	su.annotateMap (draw, model, options=ops, results=anno_results)
	del draw, ops

	#SAVE IMAGE TO DISK
	saveImage(img, model, imgName, imgDir=imgDir)



def drawProfile (model, upNode=None, dnNode=None, imgName=None, imgDir=None, width = 1024, height=512, drawSizeMultiplier = .01,
				bgroundColor=(0,3,18), bbox = None, conduitSubset = None, extraData=None):


	w = width*2
	h = height*2
	if upNode:
		#trace from up node to downstream stop node
		traced = su.traceFromNode(model, startNode=upNode, mode='down', stopnode=dnNode)
		conduitSubset = traced['conduits']
		nodeSubset = traced['nodes']

	#match the coordinates for each conduit
	conduitData = model.organizeConduitData(subset = conduitSubset)
	nodeData = model.organizeNodeData(subset=nodeSubset)
	conduits = conduitData['conduit_objects']
	nodes = nodeData['node_objects']
	minEl = nodeData['minEl']
	maxEl = nodeData['maxEl']
	reachLength = su.length_of_conduits(conduits)

	#transform
	profileHeight = maxEl - minEl

	sX = w / reachLength # to scale down from coordinate to pixels
	sY = h / profileHeight # to scale down from coordinate to pixels

	print w, " x " , h
	print 'profileHeight = ', profileHeight
	print "minEl = ",  minEl
	print "maxEl = ",  maxEl
	print "shiftRatioY = ",  sY
	print "shiftRatioX = ",  sX
	print 'reachLength = ', reachLength

	#return nodes
	img = Image.new('RGB', (w, h))
	draw = ImageDraw.Draw(img)
	x=0 #starting x location
	print '{}, {}, {}, {}, {}, {}'.format('conduit.id', 'x', 'upinvert', 'crown', 'flood_el', 'dnnode.maxHGL')
	for c in conduitSubset:

		#collect data about the curren conduit
		conduit = conduits[c]#grab object from dict\
		l = conduit.length
		upnode = nodes[conduit.upNodeID]
		dnnode = nodes[conduit.downNodeID]

		upfloodel = upnode.invert + upnode.maxDepth
		dnfloodel = dnnode.invert + dnnode.maxDepth

		inletoffset = conduit.inletoffset
		outletoffset = conduit.outletoffset
		geom1 = conduit.geom1
		upcondinv = upnode.invert + inletoffset
		dncondinv = dnnode.invert + outletoffset

		#compute the draw coordinates for each line segment
		upcondinvxy = su.coordToDrawCoord((x, upcondinv),shiftRatio=sX, shiftRatioY=sY, width=w, height=h)
		dncondinvxy = su.coordToDrawCoord((x+l, dncondinv),shiftRatio=sX, shiftRatioY=sY, width=w, height=h)
		#s = '{}: x={} upinv={}, dninv={}, uppeak={}, downpk={}'.format(conduit.id,x, upnode.invert, dnnode.invert, upnode.maxHGL, dnnode.maxHGL)
		s = '{}, {}, {}, {}, {}, {}'.format(conduit.id,x, upcondinv, upcondinv + geom1, upfloodel, upnode.maxHGL)
		x += l
		#print s
		print (upcondinvxy, dncondinvxy)
		draw.line((upcondinvxy, dncondinvxy), fill = (200, 200, 200), width = 2)



		#coordToDrawCoord(coordinates, bbox=None, shiftRatio=None, width=None, height=None)

	print '{}, {}, {}, {}, {}, {}'.format(conduit.id,x, upcondinv, upcondinv + geom1, upfloodel, upnode.maxHGL)
	saveImage(img, model, imgName)
		#draw bottom of current conduit
		#coords = None
		#dr_inv_coords = su.coordToDrawCoord(coordinates, bbox=None, shiftRatio=None, width=None, height=None)



	# for conduitID, data in conduiteElevationDicts.iteritems():
	#
	# 	#print data
	# 	drawLength = round(data['length'] * shiftRatioX, 5)
	# 	#need to shirt the x n y components of the length
	#
	# 	inv1 =		round((maxEl - data['upstreamEl']) * shiftRatioY, 5) #subtract from height because PIL origin is in top right
	# 	inv2 = 		round((maxEl - data['downstreamEl'])  * shiftRatioY, 5)  #subtract from height because PIL origin is in top right
	# 	HGL1 = 		round((maxEl - data['maxHGLUpstream']) * shiftRatioY, 5)
	# 	HGL2 = 		round((maxEl - data['maxHGLDownstream']) * shiftRatioY, 5)
	# 	upFloodEl = None
	# 	dnFloodEl = None
	# 	if extraData and data['upFloodEl'] and data['dnFloodEl']:
	# 		upFloodEl = round((data['upFloodEl']) * shiftRatioY, 5)
	# 		dnFloodEl = round((data['dnFloodEl']) * shiftRatioY, 5)
	#
	# 	tranformedDictItems = {'drawLength':drawLength, 'inv1':inv1, 'inv2':inv2, 'HGL1':HGL1, 'HGL2':HGL2, 'upFloodEl':upFloodEl, 'dnFloodEl':dnFloodEl}
	#
	#
	# 	#coordPairDict.update(tranformedDictItems)
	# 	data.update(tranformedDictItems)
	#
	#
	# print "new size X2 = " , modelSize
	# print "startingConduitID = " + startingConduitID
	# img = Image.new('RGB', (width, height), bgroundColor)
	# draw = ImageDraw.Draw(img)
	#
	# #draw conduits from the starting pipe
	# conduit = conduiteElevationDicts[startingConduitID]
	# #print conduit
	# x1 = 0
	# terminalFound = False
	#
	#
	# while not terminalFound:
	#
	# 	#loop through conduits, switching to each conduit via the "downID"
	# 	#stop when the t
	# 	length = conduit['drawLength']
	# 	inv1 = conduit['inv1']
	# 	inv2 = conduit['inv2']
	# 	HGL1 = conduit['HGL1']
	# 	HGL2 = conduit['HGL2']
	# 	upFloodEl = conduit['upFloodEl']
	# 	dnFloodEl = conduit['dnFloodEl']
	#
	# 	geom1 = float(conduit['geom1']) * shiftRatioY
	# 	x2 = length + x1 #du.getX2(inv1, inv2, length, x1)
	# 	invCoordPair = [(x1, inv1), (x2, inv2)] #coordPairs for the pipe invert
	# 	crwnCoordPair = [(x1, inv1-geom1), (x2, inv2-geom1)] #coordPairs for the pipe crown
	# 	HGLCoordPair = [(x1, HGL1), (x2, HGL2)] #coordPairs for the HGL
	# 	#print invCoordPair
	# 	draw.line(invCoordPair, fill = (200, 200, 200), width = 2)
	# 	draw.line(crwnCoordPair, fill = (200, 200, 200), width = 2)
	# 	draw.line(HGLCoordPair, fill = (10, 120, 250), width = 2)
	#
	# 	if upFloodEl and dnFloodEl:
	# 		floodElCoordPair = [(x1, inv1-upFloodEl), (x2, inv2-dnFloodEl)] #coordPairs for the pipe crown
	# 		draw.line(floodElCoordPair, fill = (30, 250, 130), width = 2)
	#
	#
	# 	nextConduitID = conduit['downID']
	# 	if nextConduitID:
	# 		#advance to next reach downstream, if a downID exists
	# 		x1 = x2 #increment the x starting point for the next conduit
	# 		#print str(conduit) + " to " + str([conduit['downID']])
	# 		conduit = conduiteElevationDicts[nextConduitID]
	# 	else:
	# 		terminalFound = True
	# 		#print "terminal" + str(conduit)
	#
	#
	# del draw
	#
	# newFile = os.path.join(imgDir, imgName) + ".png"
	# print newFile
	#
	#
	# #shrink for antialiasing niceness (though this blows up the file about 5x)
	# size = (int(width*0.5), int(height*0.5))
	# img.thumbnail(size, Image.ANTIALIAS)
	#
	# img.save(newFile)
	# os.startfile(newFile)
	# #return conduiteElevationDictss



from .run_models.run import run_simple, run_hot_start_sequence
from .run_models import start_pool
from swmmio import Model
from itertools import chain
import os
import argparse
from multiprocessing import Pool, cpu_count
from datetime import datetime

#parse the arguments
parser = argparse.ArgumentParser(description='Process some stuff')
parser.add_argument('-r', '--run', dest='model_to_run')
parser.add_argument('-rhs', '--run_hotstart', dest='hotstart_model_to_run')
parser.add_argument('-sp', '--start_pool', dest='start_pool', nargs="+")
parser.add_argument('-cores_left', '--cores_left', dest='cores_left', default=4, type=int)

args = parser.parse_args()
wd = os.getcwd() #current directory script is being called from




if args.model_to_run is not None:
    print 'run the model: {}'.format(args.model_to_run)
    run_simple(args.model_to_run)

elif args.hotstart_model_to_run is not None:
    print 'hotstart_model_to_run the model: {}'.format(args.hotstart_model_to_run)
    m = Model(args.hotstart_model_to_run)
    run_hot_start_sequence(m)#args.hotstart_model_to_run)

elif args.start_pool is not None:

    models_dirs = [os.path.join(wd, f) for f in args.start_pool]
    print 'looking for models to queue in:\n\t{}'.format('\n\t'.join(models_dirs))
    #combine the segments and options (combinations) into one iterable
    # paths = (segments_dir, options_dir)
    # baseline = Model(baseline_dir)
    dirs_containing_inps = []
    for root, dirs, files in chain.from_iterable(os.walk(path) for path in models_dirs):
        for file in files:
            if file.endswith('.inp') and 'bk' not in root:
                #we've found a directory containing an inp
                print file
                dirs_containing_inps.append(root)


    start_pool.main(dirs_containing_inps, args.cores_left)
    # pool = Pool(cpu_count() - args.cores_left) # use all but 6 available cores
    # proces data_inputs iterable with pool
    # print pool
    # res = pool.map(run_swmm_engine, dirs_containing_inps)
    # map(run_swmm_engine, dirs_containing_inps[0:2])
    print "hi im done running"
    # baseline_dir = r'F:\models\SPhila\MasterModels_170104\Baseline' #r'P:\06_Tools\v_control\BIG\Baseline'
    # log_dir = r'F:\models\SPhila\MasterModels_170104\ProjectAdmin'
    # batch.batch_post_process(combi_folder, baseline_dir, log_dir, bbox=su.d68d70)


    print 'start_pool: {}\n{} cores left'.format(args.start_pool,
                                                 args.cores_left)


else:
    print 'you need to pass in some args'



#!/usr/bin/env python
#coding:utf-8

#this is fairly specific stuff and will only work given certain geospatial data
from .utils import swmm_utils as su
import os
import pickle
import json
from definitions import *

class Parcel(object):

	#object representing a swmm node object

	__slots__ = ('id', 'flood_duration', 'avg_flood_duration',
				'nodes', 'sheds', 'delta_type', 'is_delta')

	def __init__(self, id, nodes=[], sheds=[]):

		#assign the header list
		self.id = id
		self.flood_duration = 0
		self.avg_flood_duration = 0
		self.nodes = nodes
		self.sheds = sheds
		self.is_delta = False
		self.delta_type = 'flooding_unchanged' #whether this object's data represents a change between two parent models

def compareParcels(existing_elements, proposed_elements, parcel_features=PARCEL_FEATURES, bbox=None,
					floodthreshold=0.08333, delta_threshold=0.25, anno_results={}):

	#return a dict of node objects with the delta between models

	# existing_elements = parcel_flood_duration(model1, parcel_features=parcel_features,
	# 										bbox=bbox, anno_results=anno_results)['parcels']
	# proposed_elements = parcel_flood_duration(model2, parcel_features=parcel_features,
	# 										bbox=bbox, anno_results=anno_results)['parcels']


	#WARNING THIS CHANGES THE VALUES IN proposed_elements


	newflood = moreflood = floodEliminated = floodlowered = 0 #for documenting how things changed
	for id, proposed in proposed_elements.iteritems():

		existing = existing_elements.get(id, {})

		if id in existing_elements:

			flood_duration_delta = proposed.flood_duration - existing.flood_duration

			if existing.flood_duration >= floodthreshold and proposed.flood_duration >= floodthreshold:
				#parcel still floods, check how it changed:
				if flood_duration_delta > delta_threshold:
					#flooding duration increased (more than delta_threhold)
					proposed.delta_type = 'increased_flooding'
					moreflood += 1

				elif flood_duration_delta < -delta_threshold:
					#flooding duration decreased (more than delta_threhold)
					proposed.delta_type = 'decreased_flooding'
					floodlowered += 1

			elif existing.flood_duration < floodthreshold and proposed.flood_duration >= floodthreshold and abs(flood_duration_delta) >= delta_threshold:
				#flooding occurs where it perviously did not
				proposed.delta_type = 'new_flooding'
				newflood += 1

			elif existing.flood_duration >= floodthreshold and proposed.flood_duration < floodthreshold and abs(flood_duration_delta) >= delta_threshold:
				#parcel that previously flooded no longer does
				proposed.delta_type = 'eliminated_flooding'
				floodEliminated += 1


			proposed.is_delta = True
			proposed.flood_duration = flood_duration_delta
			proposed.avg_flood_duration -= existing.avg_flood_duration


	#update annotation results
	impact_dict = {
				1:{'title':'New flooding','val':newflood},
				2:{'title':'Increased flooding','val':moreflood},
				3:{'title':'Decreased flooding','val':floodlowered},
				4:{'title':'Eliminated flooding','val':floodEliminated},
				}
	impacts_string = "\n"
	for i in sorted(impact_dict):
		title = impact_dict[i]['title']
		val = impact_dict[i]['val']
		impacts_string += '{} on {} parcels\n'.format(title, val)

	anno_title =  '\nImpact Of Alternative (dt >= {}m)'.format(int(round(delta_threshold*60)))
	anno_results.update({anno_title:impacts_string})
	results = {
				'parcels':proposed_elements,
				'increased_flooding':moreflood,
				'decreased_flooding':floodlowered,
				'new_flooding': newflood,
				'eliminated_flooding':floodEliminated
			}
	return results

def parcel_flood_duration(model, parcel_features, threshold=0.083,  bbox=None,
							gdb=GEODATABASE,
							export_table=False,anno_results={}):
	"""
	this method includes the logic for calculating a flood duration (and avg
	duration) on the given parcels. Returned is a dictionary with each parcel ID
	and a Parcel object with the flooding data populated.
	"""

	#collect the parcels in a dictionary of Parcel objects
	parcel_objects = associate_parcels(model, parcel_features, gdb=gdb, bbox=bbox)

	#grab the list of flooded nodes, and create the dictionary of parcels linked to node IDs
	flooded_nodes = su.subsetElements(model, min=threshold, bbox=bbox, pair_only=True)

	#tally how many were flooded above set durations
	#{minutes : number of parcels with flooding greater than}
	duration_partition = {
				5:0, 10:0, 15:0,
				30:0, 60:0, 120:0
				}
	#calculate average flood duration for each parcel
	parcels_flooded_count = 0

	for id, parcel in parcel_objects.iteritems():

		if len(parcel.nodes) > 0:
			#parcel has nodes associated

			total_parcel_flood_dur = 0.0
			flood_duration = 0.0
			for node in parcel.nodes:

				#look up the flood duration
				node_duration = flooded_nodes.get(node, 0)
				total_parcel_flood_dur += node_duration #for avereage calculation

				#parcel flooding duration assumed to be the max of all adjacent node durations
				flood_duration = max(flood_duration, node_duration)

			avg_flood_duration = total_parcel_flood_dur/len(parcel.nodes)

			parcel.avg_flood_duration = avg_flood_duration
			parcel.flood_duration = flood_duration

			if flood_duration >= threshold:
				#we've found a parcel that is considered flooded
				parcels_flooded_count += 1

			for duration, count in duration_partition.iteritems():
				if flood_duration >= float(duration)/60.0:
					count += 1
					duration_partition.update({duration:count})

	parcels_count = len(parcel_objects)
	parcels_flooded_fraction = float(parcels_flooded_count)/float(parcels_count)

	results = {
				'parcels_flooded_count':parcels_flooded_count,
				'parcels_count':parcels_count,
				'parcels_flooded_fraction':parcels_flooded_fraction,
				'duration_partition':duration_partition
				}

	results_string = "{} ({}%) of {} total".format(parcels_flooded_count, round(parcels_flooded_fraction*100),parcels_count)

	#print results_string

	#partition (detailed) results string
	partitioned_results = "\n"
	for d in sorted(duration_partition):
		perc_of_tot = int(round( float(duration_partition[d]) / float(results['parcels_count']) * 100 ))
		partitioned_results += '>{}mins : {} ({}%)\n'.format(d, duration_partition[d], perc_of_tot)

	#track results for annotation
	anno_results.update({'Total Parcels':results['parcels_count'], '\nParcels Flooded':partitioned_results})

	#add in the actual list of parcels for drawing
	results.update({'parcels':parcel_objects})

	return results

def associate_parcels(model, feature=PARCEL_FEATURES,
						cols = ["PARCELID", "OUTLET", "SUBCATCH", "SHAPE@"],
						bbox=None, gdb=GEODATABASE):

	"""
	create dictionary with keys for each parcel, and sub array containing
	associated nodes. This method expects a shapefile that results from a spatial
	join (one to many) between the given model drainage areas and a parcels
	spatial dataset in the study area.

	Where parcels spatially fall within multiple sheds, they will be
	represented in multiple rows in this shapefile (one to many), with one row
	for each associated shed. given this, this function creates a list of unique
	parcels and arrays with their associated sheds
	"""

	#check if a parcel to node association dicitonary exists, load if possible
	parcel_to_nodes_filename = os.path.join(model.inp.dir, 'parcel_nodes_dict.txt')
	if not os.path.isfile(parcel_to_nodes_filename):
		print 'creating new parcel/node association dict'
		#build the dictionary of parcels, if this hasn't already been done
		features = os.path.join(gdb, feature)
		import arcpy
		parcels_dictionary = {}
		for row in arcpy.da.SearchCursor(features, cols):

			#first check if parcel is in bbox
			jsonkey = 'rings' #for arc polygons
			geometrySections = json.loads(row[3].JSON)[jsonkey]
			parcel_in_bbox=True #assume yes first
			for i, section in enumerate(geometrySections):
				#check if part of geometry is within the bbox, skip if not
				if bbox and len ( [x for x in section if su.pointIsInBox(bbox, x)] ) == 0:
					parcel_in_bbox=False #continue #skip this section if none of it is within the bounding box

			if not parcel_in_bbox:
				continue #skip if not in bbox

			PARCELID = str(row[0])
			if PARCELID in parcels_dictionary:
				#append to existing parcel object's arrays
				parcels_dictionary[PARCELID]['nodes'].append(row[1])
				parcels_dictionary[PARCELID]['sheds'].append(row[2])
			else:
				#new parcel id found. instantiate new object with initial
				#associated nodes and sheds
				#parcel = Parcel(PARCELID, nodes=[row[1]], sheds=[row[2]])
				parcels_dictionary.update({PARCELID:{'nodes':[row[1]], 'sheds':[row[2]]}}  )

		#save for later use
		#print 'writing new parcel/node association dict'
		with open(parcel_to_nodes_filename, 'w') as dictSaveFile:
			pickle.dump(parcels_dictionary, dictSaveFile)

	else:
		#print "loading parcel/node association dict: {}".format(parcel_to_nodes_filename)
		parcels_dictionary = pickle.load( open(parcel_to_nodes_filename, 'r') )
		#print "WARNING: make sure this loaded parcel dict contains all parcels!"


	#create a dictionary of parcel objects
	parcels = {}
	for id, parcel_data in parcels_dictionary.iteritems():
		parcel = Parcel(id, nodes = parcel_data['nodes'], sheds = parcel_data['sheds'])
		parcels.update({id:parcel})


	return parcels



#UTILITY FUNCTIONS AIMED AT I/O OPERATIONS WITH TEXT FILES
#STANDARD READING AND WRITING OF TEXT FILES (E.G. .INP AND .RPT)

import os
from swmmio.utils.functions import random_alphanumeric, complete_inp_headers, complete_rpt_headers


#default file path suffix
txt = '.txt'

def inline_comments_in_inp(filepath, overwrite=False):

    """
    with an existing INP file, shift any comments that have been placed above the
    element (behavoir from saving in GUI) and place them to the right, inline with
    the element. To improve readability
    """
    newfilename = os.path.splitext(os.path.basename(filepath))[0] + '_unGUI.inp'
    newfilepath = os.path.join(os.path.dirname(filepath), newfilename)
    allheaders = complete_inp_headers(filepath)

    with open(filepath) as oldf:
        with open(newfilepath, 'w') as new:

            comment_concat = [] #to hold list of comments (handles multiline comments)
            current_section = allheaders['order'][0]
            for line in oldf:

                #determine what section we are in by noting when we pass double brackets
                if '[' and ']' in line:
                    current_section = line.strip()


                #if line.count(';') == 1 and line.strip()[0]==';':
                if len(line.strip()) > 1:
                    if line.strip()[0]==';' and ''.join(line.strip()[:2]) != ';;':
                        #print line.strip()[:2]
                        #this is a comment bc first char is ; and there
                        #seconf char is not (which would resember the header section)
                        words = line.split()
                        hdrs = allheaders['headers'][current_section].split()#headerrow.split()
                        perc_match_to_header = float(len([x for x in words if x in hdrs])) / float(len(hdrs))
                        if perc_match_to_header <= 0.75:
                            comment_concat.append(line.strip())
                    else:
                        #print comment_concat
                        #this row has data, tack any comment to the line end
                        comment_string = ''
                        if len(comment_concat) > 0:
                            comment_string = r' '.join(comment_concat)
                        newlinestring = line.strip() + comment_string + '\n'
                        new.write(newlinestring)
                        comment_concat = []
                else:
                    #write the short line
                    new.write(line)

    #rename files and remove old if we should overwrite
    if overwrite:
        os.remove(filepath)
        os.rename(newfilepath, filepath)



def extract_section_from_file(filepath, sectionheader, element_id=None, cleanheaders=True, startfile=False, headerdefs=None,  ignore_comments=False):
    f_extension = os.path.splitext(filepath)[1]
    if f_extension == '.inp':
        return extract_section_from_inp(filepath, sectionheader, cleanheaders, startfile, headerdefs, ignore_comments)
    if f_extension == '.rpt':
        return extract_section_from_rpt(filepath, sectionheader, element_id, cleanheaders, startfile, headerdefs)

def extract_section_from_inp(filepath, sectionheader, cleanheaders=True,
                            startfile=False, headerdefs=None, ignore_comments=False,
                            return_string=False, skiprows=0, skipheaders=False):
    """
    INPUT path to text file (inp, rpt, etc) and a text string
    matchig the section header of the to be extracted

    creates a new text file in the same directory as the filepath
    and returns the path to the new file.

    optionally inserts row at begining of file with one-line headers that are
    defined in swmmio.defs.sectionheaders (useful for creating dataframes)
    """

    #headers = she.complete_inp_headers(inp)['headers']
    if not headerdefs:
        allheaders = complete_inp_headers(filepath)['headers']
    else:
        allheaders = headerdefs['headers']

    with open(filepath) as f:
        startfound = False
        endfound = False
        #outFilePath = inp.dir + "\\" + inp.name + "_" + section + ".txt"
        wd = os.path.dirname(filepath)
        newfname = sectionheader + '_' + random_alphanumeric(6)
        outfilepath = os.path.join(wd, newfname + txt)
        out_string = '' #populate if we only want a string returned
        line_count = 0
        with open(outfilepath, 'w') as newf:

            for line in f:

                if startfound and line.strip() in allheaders:
                    endfound = True
                    #print 'end found: {}'.format(line.strip())
                    break
                elif not startfound and sectionheader in line:
                    startfound = True
                    #replace line with usable headers
                    if cleanheaders:
                        if allheaders[sectionheader] != 'blob':
                            line = allheaders[sectionheader] + '\n'
                        else:
                            line = sectionheader + '\n'

                if startfound:
                    if ignore_comments:
                        #ignore anything after a comment
                        #THIS IS HACK GARBAGE
                        if ';' in line:
                            s = line.split(';')[0] + '\n'
                            out_string += s #build the overall out string (optional use)
                            newf.write(s)
                        else:
                            newf.write(line)
                            out_string += line
                    elif skipheaders:
                        #check if we're at a inp header row with ';;' or the section
                        #header e.g. [XSECTIONS]. If so, skip the row bc skipheader = True
                        if line.strip()[:2] != ";;" and line.strip() != sectionheader:
                            newf.write(line)
                            out_string += line
                    elif line_count >= skiprows:
                        newf.write(line)
                        out_string += line

                    line_count += 1

    if not startfound:
        #input section header was not found in the file
        os.remove(outfilepath)
        return None
    if return_string:
        #return a string rep of the extracted section
        os.remove(outfilepath) #not needed so delete (maybe wasteful to make in 1st place)
        return out_string

    if startfile:
        os.startfile(outfilepath)

    return outfilepath


def extract_section_from_rpt(filepath, sectionheader, element_id=None, cleanheaders=True, startfile=False, headerdefs=None):
    """
    INPUT path to rpt file and a text string
    matchig the section header of the to be extracted

    creates a new text file in the same directory as the filepath
    and returns the path to the new file.

    optionally inserts row at begining of file with one-line headers that are
    defined in swmmio.defs.sectionheaders (useful for creating dataframes)
    dede
    """
    if not headerdefs:
        allheaders = complete_rpt_headers(filepath)['headers']
    else:
        allheaders = headerdefs['headers']

    with open(filepath) as f:


        wd = os.path.dirname(filepath)
        newfname = sectionheader + '_' + random_alphanumeric(6)
        outfilepath = os.path.join(wd, newfname + txt)

        sectionstartfound = False #where we see the section header name
        elementstartfound = False #where we see the element time series
        datastartfound = False #where we see actual data after the header text
        endfound = False
        hyphencount = 0 #count how many hyphen rows we pass after sectionstartfound

        with open(outfilepath, 'w') as newf:

            for line in f:

                if sectionstartfound and line.strip() in allheaders:
                    #if we've found the begining of the section and we run
                    #into another known header, we're at the end
                    endfound = True
                    break

                elif not sectionstartfound and sectionheader in line:
                    #if we haven't found the start of the section yet, the
                    #sectionheader is in the current line,
                    #then we've found the begining of the section of interest
                    #print '{} section found'.format(sectionheader)
                    sectionstartfound = True

                    #replace line with usable headers
                    if cleanheaders and [sectionheader] != 'blob':
                        headerrow = allheaders[sectionheader] + '\n'
                        newf.write(headerrow)

                if sectionstartfound and element_id:
                    #keep searching for the start of actual good data
                    #string starting an element time series e.g. <<< Node D68_1_4 >>>
                    elem_start_string = ' '.join(["<<<", sectionheader.split()[0], element_id, ">>>"])

                    if element_id and elem_start_string in line:
                        print 'element_id found: {}'.format(line)
                        #if we should look for an element_id and it
                        #is in the current line
                        elementstartfound = True
                        #datastartfound = True
                    if elementstartfound:
                        if hyphencount == 2:
                            datastartfound = True
                            newf.write(line) #write the first line
                            if len(line.strip()) == 0:
                                #if line is blank after we've found the data,
                                #we can be confident that we've found the end
                                break
                        if '------------' in line:
                            #track how many rows with hyphens are passed
                            #two rows should occur before actual data is written
                            hyphencount += 1
                elif sectionstartfound and not element_id:

                    if hyphencount == 2:
                        datastartfound = True
                        newf.write(line) #write the first line
                        if len(line.strip()) == 0:
                            #if line is blank after we've found the data,
                            #we can be confident that we've found the end
                            break
                    if '------------' in line:
                        #track how many rows with hyphens are passed
                        #two rows should occur before actual data is written
                        hyphencount += 1

    if not sectionstartfound:
        #input section header was not found in the file
        return None
    if startfile:
        os.startfile(outfilepath)

    return outfilepath
    #pass



import subprocess
import os
import pandas as pd
from swmmio.version_control import version_control as vc
from swmmio.utils.modify_model import replace_inp_section
from swmmio.utils import dataframes
from swmmio.run_models import defs
from definitions import SWMM_ENGINE_PATH


#path to the SWMM5 Engine
# SWMM_ENGINE_PATH = r'\\PWDHQR\Data\Planning & Research\Flood Risk Management\07_Software\swmm5_22.exe'

def run_simple(inp_path, swmm_eng=SWMM_ENGINE_PATH):
    """
    run a model once as is.
    """
    print 'running {} with {}'.format(inp_path, swmm_eng)
    #inp_path = model.inp.filePath
    rpt_path = os.path.splitext(inp_path)[0] + '.rpt'

    subprocess.call([swmm_eng, inp_path, rpt_path])

def run_hot_start_sequence(model, swmm_eng=SWMM_ENGINE_PATH):

    inp_path = model.inp.filePath
    rpt_path = os.path.splitext(inp_path)[0] + '.rpt'
    hotstart1 = os.path.join(model.inp.dir, model.inp.name + '_hot1.hsf')
    hotstart2 = os.path.join(model.inp.dir, model.inp.name + '_hot2.hsf')

    if not os.path.exists(hotstart1) and not os.path.exists(hotstart2):
        #create new model inp with params to save hotstart1
        print 'create new model inp with params to save hotstart1'
        s = pd.Series(['SAVE HOTSTART "{}"'.format(hotstart1)])
        hot1_df = pd.DataFrame(s, columns=['[FILES]'])
        model = replace_inp_section(model.inp.filePath, '[FILES]', hot1_df)
        model = replace_inp_section(model.inp.filePath, '[REPORT]', defs.REPORT_none)
        model = replace_inp_section(model.inp.filePath, '[OPTIONS]', defs.OPTIONS_no_rain)
        subprocess.call([swmm_eng, model.inp.filePath, rpt_path])

    if os.path.exists(hotstart1) and not os.path.exists(hotstart2):
        #create new model inp with params to use hotstart1 and save hotstart2
        print 'with params to use hotstart1 and save hotstart2'
        s = pd.Series(['USE HOTSTART "{}"'.format(hotstart1), 'SAVE HOTSTART "{}"'.format(hotstart2)])
        hot2_df = pd.DataFrame(s, columns=['[FILES]'])
        model = replace_inp_section(model.inp.filePath, '[FILES]', hot2_df)
        subprocess.call([swmm_eng, model.inp.filePath, rpt_path])

    if os.path.exists(hotstart2):
        #create new model inp with params to use hotstart2 and not save anything
        print 'params to use hotstart2 and not save anything'
        s = pd.Series(['USE HOTSTART "{}"'.format(hotstart2)])
        hot3_df = pd.DataFrame(s, columns=['[FILES]'])

        model = replace_inp_section(model.inp.filePath, '[FILES]', hot3_df)
        model = replace_inp_section(model.inp.filePath, '[REPORT]', defs.REPORT_none)# defs.REPORT_nodes_links)
        model = replace_inp_section(model.inp.filePath, '[OPTIONS]', defs.OPTIONS_normal)

        subprocess.call([swmm_eng, model.inp.filePath, rpt_path])



from swmmio.utils import text as txt
from swmmio.utils import functions as funcs
from swmmio.swmmio import rpt, inp
import pandas as pd
import os

def create_dataframeBI(bi_path, section='[CONDUITS]'):
    """
    given a path to a biuld instructions file, create a dataframe of data in a
    given section
    """
    headerdefs = funcs.complete_inp_headers(bi_path)
    headerlist = headerdefs['headers'][section].split() + [';', 'Comment', 'Origin']
    tempfilepath = txt.extract_section_from_inp(bi_path, section,
                                                headerdefs=headerdefs,
                                                skipheaders=True)
    df = pd.read_table(tempfilepath, header=None, delim_whitespace=True,
                       skiprows=[0], index_col=0, names = headerlist, comment=None)

    os.remove(tempfilepath) #clean up

    return df

def create_dataframeINP (inp, section='[CONDUITS]', ignore_comments=True, comment_str=';'):
    """
    given an inp object, create a dataframe of data in a given section
    """

    #handle an inp object or a path to an INP file
    if type(inp) is not str:
        inp_path = inp.filePath
    else:
        inp_path = inp

    #find all the headers and their defs (section title with cleaned one-liner column headers)
    headerdefs = funcs.complete_inp_headers(inp_path)
    #create temp file with section isolated from inp file
    tempfilepath = txt.extract_section_from_inp(inp_path, section,
                                                headerdefs=headerdefs,
                                                ignore_comments=ignore_comments)
    if ignore_comments:
        comment_str = None
    if not tempfilepath:
        #if this head (section) was not found in the textfile, return a
        #blank dataframe with the appropriate schema
        print 'header "{}" not found in "{}"'.format(section, inp_path)
        print 'returning empty dataframe'
        headerlist = headerdefs['headers'].get(section, 'blob').split() + [';', 'Comment', 'Origin']
        blank_df = pd.DataFrame(data=None, columns=headerlist).set_index(headerlist[0])
        return blank_df

    if headerdefs['headers'][section] == 'blob':
        #return the whole row, without specifc col headers
        df = pd.read_table(tempfilepath, delim_whitespace=False, comment=comment_str)
    elif section == '[CURVES]' or section =='[TIMESERIES]':
        #return the whole row, without specifc col headers
        df = pd.read_table(tempfilepath, delim_whitespace=False)#, index_col=0)#, skiprows=[0])
    else:
        #this section header is recognized and will be organized into known columns
        headerlist = headerdefs['headers'][section].split() + [';', 'Comment', 'Origin']
        df = pd.read_table(tempfilepath, header=None, delim_whitespace=True, skiprows=[0],
                             index_col=0, names = headerlist, comment=comment_str)
        df[';'] = ';'
        #df.columns.names=[headerlist]
    os.remove(tempfilepath)

    #add new blank comment column after a semicolon column
    # df[';'] = ';'
    # df['Comment'] = ''


    return df

def create_dataframeRPT(rpt, section='Link Flow Summary', element_id=None):
    """
    given an rpt object, create a dataframe of data in a given section
    """

    #find all the headers and their defs (section title with cleaned one-liner column headers)
    headerdefs = funcs.complete_rpt_headers(rpt.filePath)
    #create temp file with section isolated from rpt file
    tempfilepath = txt.extract_section_from_rpt(rpt.filePath, section, headerdefs=headerdefs, element_id=element_id)

    if not tempfilepath:
        print 'header "{}" not found in "{}"'.format(section, rpt.filePath)
        return None

    if headerdefs['headers'][section] == 'blob':
        #return the whole row, without specifc col headers
        df = pd.read_table(tempfilepath, delim_whitespace=False, comment=";")
    else:
        if element_id:
            #we'retrying to pull a time series, parse the datetimes by
            #concatenating the Date Time columns (cols 1,2)
            df0 = pd.read_table(tempfilepath, delim_whitespace=True)
            df = df0[df0.columns[2:]] #the data sans date time columns
            df.index=pd.to_datetime(df0['Date'] + ' ' + df0['Time'])
            df.index.name = "".join(df0.columns[:2])
        else:
            #this section header is recognized and will be organized into known columns
            #headerlist = headerdefs['headers'][section].split()
            df = pd.read_table(tempfilepath, delim_whitespace=True, index_col=0)
        #df.columns.names=[headerlist]
    os.remove(tempfilepath)

    #add new blank comment column
    #df['Comment'] = ''


    return df



"""
Utility functions related to comparing SWMM models
"""
import pandas as pd
from swmmio.version_control import inp
from swmmio.utils.dataframes import create_dataframeINP
import os

def length_of_new_and_replaced_conduit(model1, model2):

    changes = inp.Change(model1, model2, section ='[CONDUITS]')
    df = pd.concat([changes.added, changes.altered])

    return df.Length.sum()

def create_shapefile_of_new_conduits(model1, model2, filename=None):
    """
    given a baseline model and a option, create a shapefile including only the
    new conduits.
    """

    changes = inp.Change(model1, model2, section='[CONDUITS]')
    df = pd.concat([changes.added, changes.altered])
    new_conduit_ids = df.index.tolist()

    if filename is None:
        filename = os.path.join(os.path.dirname(model2.inp.filePath),
                                'shapefiles',
                                model2.inp.name + '_new_conduits.shp')

    #create the containing directory if necessary
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))

    model2.export_to_shapefile(element_type='conduit',
                               filename=filename,
                               subset=new_conduit_ids)




def estimate_cost_of_new_conduits(baseline, newmodel, additional_costs=None):

    changes = inp.Change(baseline, newmodel, section ='[CONDUITS]')
    newconduits = pd.concat([changes.added, changes.altered])
    newconduits.drop([';', 'Comment', 'Origin'], axis=1, inplace=True)

    xsections = create_dataframeINP(newmodel.inp, section='[XSECTIONS]')
    xsections.drop([';', 'Comment', 'Origin'], axis=1, inplace=True)

    #join the xsection data to the newconduits df. convert geoms to numbers
    newconduits = newconduits.join(xsections)
    geoms = ['Geom1','Geom2','Geom2','Geom4']
    newconduits[geoms] = newconduits[geoms].apply(pd.to_numeric)

    def calc_area(row):
        """calculate the cross-sectional area of a sewer segment"""

        if row.Shape == 'CIRCULAR':
            d = row.Geom1
            area = 3.1415 * (d * d) / 4
            return round((area * row.Barrels),2)

        if 'RECT' in row.Shape:
            #assume triangular bottom sections (geom3) deepens the excavated box
            return (row.Geom1 + row.Geom3) * row.Geom2
        if row.Shape =='EGG':
            #assume geom1 is the span
            return row.Geom1*1.5

    def get_unit_cost(row):
        cost_dict = {1.0:570,
                    1.5:570,
                    1.75:610,
                    2.0:680,
                    2.25:760,
                    2.5:860,
                    3.0:1020,
                    3.5:1200,
                    4.0:1400,
                    4.5:1550,
                    5.0:1700,
                    5.5:1960,
                    6.0:2260,
                    7.0:2600,
                    7.5:3000,
                    }

        def round_to(n, precision):
            correction = 0.5 if n >= 0 else -0.5
            return int( n/precision+correction ) * precision

        def round_to_05(n):
            return round_to(n, 0.05)

        cleaned_geom = round_to_05(row.Geom1)
        # print '{}: Raw Geom1 = {} is rounded  to {}'.format(row.InletNode, row.Geom1, cleaned_geom)

        RectBox_UnitCost = 80
        if row.Shape == 'CIRCULAR':
            try:
                val = cost_dict[cleaned_geom]
            except:
                val = 0
            return val
        if 'RECT' in row.Shape:
            """
            assume any triangular bottom section adds to
            overall excavation box section
            """
            val = round(RectBox_UnitCost*row.XArea,2)
            return val
        if row.Shape == 'EGG':
            """
            We're going to treat this like a box"
            """
            val = round(RectBox_UnitCost*row.XArea,2)
            return val


    def compute_conduit_cost(row):
        return row.UnitCostLF * row.Length

    def compute_volume(row):
        return row.XArea * row.Length

    def added_cost(row):
        #add in any additional cost data (from crossing water mains, etc)
        return row.CostEstimate + row.AdditionalCost

    newconduits['XArea'] = newconduits.apply (lambda row: calc_area (row), axis=1)
    newconduits['UnitCostLF'] = newconduits.apply(lambda row: get_unit_cost(row), axis=1)
    newconduits['Volume'] = newconduits.apply (lambda row:
                                               compute_volume (row), axis=1)
    newconduits['CostEstimate'] = newconduits.apply (lambda row:
                                                     compute_conduit_cost (row),
                                                     axis=1)

    if additional_costs:
        #read in the supplemental cost data from the csv
        addcosts = pd.read_csv(additional_costs, index_col=0)
        newconduits = newconduits.join(addcosts).fillna(0)
        newconduits['TotalCostEstimate'] = newconduits.apply (lambda row:
                                                              added_cost(row),
                                                              axis=1)
    else:
        # NOTE this lingo here is weak...
        #additional_costs not provided, rename the CostEstimate to TotalCostEstimate
        newconduits = newconduits.rename(columns={"CostEstimate": "TotalCostEstimate"})

    return newconduits



#!/usr/bin/env python
#coding:utf-8

import random
from time import gmtime, strftime
import re
import os
import numpy
import pandas as pd
import parcels
import pickle
from .utils import swmm_utils as su
import glob
import csv
from .utils import text as txt

class Model(object):

	#Class representing a complete SWMM model incorporating its INP and RPT
	#files and data

	def __init__(self, in_file_path):

		#can init with a directory containing files, or the specific inp file
		"""
		initialize a swmmio.Model object by pointing it to a directory containing
		a single INP (and optionally an RPT file with matching filename) or by
		pointing it directly to an .inp file.
		"""

		inp_path = None
		if os.path.isdir(in_file_path):
			#a directory was passed in
			#print 'is dir = {}'.format(in_file_path)
			inps_in_dir = glob.glob1(in_file_path, "*.inp")
			if len(inps_in_dir) == 1:
				#there is only one INP in this directory -> good.
				inp_path = os.path.join(in_file_path, inps_in_dir[0])
				#print 'only 1 inp found = {}'.format(inp_path)

		elif os.path.splitext(in_file_path)[1] == '.inp':
			#an inp was passed in
			inp_path = in_file_path
			#print 'is inp path = {}'.format(in_file_path)

		if inp_path:
			wd = os.path.dirname(inp_path) #working dir
			name = os.path.splitext(os.path.basename(inp_path))[0] #basename
			self.name = name
			self.inp = inp(inp_path) #inp object
			self.rpt = None #until we can confirm it initializes properly
			#slots to hold processed data
			self.organized_node_data = None
			self.organized_conduit_data = None
			self.bbox = None #to remember how the model data was clipped

			#try to initialize a companion RPT object
			rpt_path = os.path.join(wd, name + '.rpt')
			if os.path.exists(rpt_path):
				try:
					self.rpt = rpt(rpt_path)
				except:
					print '{}.rpt failed to initialize'.format(name)


	def organizeConduitData (self, bbox=None, subset=None, extraData=None, findOrder=False):

		#creates a dictionary of dictionaries containing relevant data
		#for conduits within a SWMM model. Upstream and downstream node data is attributed to the

		#check if this has been done already and return that data accordingly
		if self.organized_conduit_data and bbox==self.bbox:
			return self.organized_conduit_data

		#parse out the main objects of this model
		inp = self.inp
		rpt = self.rpt

		#creat dictionary of conduits and join dictionaries of all nodes
		conduitsDict = inp.createDictionary("[CONDUITS]")
		xsectionsDict = inp.createDictionary("[XSECTIONS]")
		if subset:
			conduitsDict = { key:value for key,value in conduitsDict.items() if key in subset }
			xsectionsDict = { key:value for key,value in xsectionsDict.items() if key in subset }
		storagesDict = inp.createDictionary("[STORAGE]")
		junctionsDict = inp.createDictionary("[JUNCTIONS]")
		outfallsDict = inp.createDictionary("[OUTFALLS]")
		coordsDict = inp.createDictionary("[COORDINATES]")

		#merge all the "node" type dictionaries (those that can be junctions between SWMM links)
		allNodesDict = su.merge_dicts(storagesDict, junctionsDict, outfallsDict)

		nodesDepthSummaryDict = {} # from rpt if one is supplied
		conduit_objects = {}
		#downstreamConduits = []
		if rpt:
			#create a dictionary holding data from an rpt file, if provided
			nodesDepthSummaryDict = rpt.createDictionary("Node Depth Summary")
			allLinksSummaryDict = rpt.createDictionary("Link Flow Summary")
			nodesFloodSummaryDict = rpt.createDictionary("Node Flooding Summary")
			#print "created node and link summary dicts from " + rpt.fName

		reachLength = 0.00 #used for tranform
		maxEl = 0.00 #used for tranform
		minEl = 999999 #used for transform
		#errCount = 0

		for conduit_id, conduit_data in conduitsDict.iteritems():

			#try:
			upstreamNodeID = conduit_data[0]
			downstreamNodeID = conduit_data[1]
			inletoffset = float(conduit_data[4])
			outletoffset = float(conduit_data[5])
			length = float(conduit_data[2])
			upstreamXY = coordsDict[upstreamNodeID]
			upstreamXY = [float(i) for i in upstreamXY] #convert to floats
			downstreamXY = coordsDict[downstreamNodeID]
			downstreamXY = [float(i) for i in downstreamXY] #convert to floats
			geom1 = float(xsectionsDict[conduit_id][1]) #conduit diameter / height
			geom2 = xsectionsDict[conduit_id][2] #conduit width (sometimes a curve ref)

			if bbox and (not su.pointIsInBox(bbox, upstreamXY) and not su.pointIsInBox(bbox, downstreamXY)):
				#skip conduits who are not within a given boudning box. This includes conduits who are partially in the box.
				continue

			#downstream pipe id: the pipe whose upstream id equals downstreamNodeID
			conduit = Link(conduit_id, [upstreamXY, downstreamXY], geom1, geom2, inletoffset, outletoffset)
			conduit.length = length
			conduit.upNodeID = upstreamNodeID
			conduit.downNodeID = downstreamNodeID

			if rpt and (upstreamNodeID in nodesDepthSummaryDict) and (downstreamNodeID in nodesDepthSummaryDict):
				# #if the up and downstream nodes are also found in the node depth summary dictionary,
				# #grab the relevant data and store a little dictionary
				upNDA = nodesDepthSummaryDict[upstreamNodeID] #upstream node depth array
				dnNDA = nodesDepthSummaryDict[downstreamNodeID] #downstream node depth array
				upHGL = float(upNDA[3])
				dnHGL = float(dnNDA[3])
				conduit.maxHGLDownstream = dnHGL
				conduit.maxHGLUpstream = upHGL


			if rpt and conduit_id in allLinksSummaryDict:
				#if the conduit is found in the link summary dictionary,
				#grab the link flow smmary data and store a little dictionary
				lsa = allLinksSummaryDict[conduit_id] #link summary array
				#lsa = [float(i) for i in lsa] #convert to floats
				if lsa[0] == "CONDUIT":

					conduit.maxflow = float(lsa[1])
					conduit.maxQpercent = float(lsa[5])


			#append data to the overall dictionary
			conduit_objects.update({conduit_id:conduit})

			#except Exception (e):
			#	print str(e)

		output = {'conduit_objects': conduit_objects}

		#remember this stuff for later
		self.organized_conduit_data = output
		self.bbox = bbox

		return output

	def organizeNodeData (self, bbox=None, subset=None):

		#creates a dictionary of dictionaries containing relevant data
		#per node within a SWMM model.

		#check if this has been done already and return that data accordingly
		if self.organized_node_data and bbox==self.bbox:
			#print "loaded node objs from " + self.rpt.fName
			return self.organized_node_data

		#parse out the main objects of this model
		inp = self.inp
		rpt = self.rpt

		#creat dictionary of all nodes from the INP and join dictionaries
		storagesDict = inp.createDictionary("[STORAGE]")
		junctionsDict = inp.createDictionary("[JUNCTIONS]")
		outfallsDict = inp.createDictionary("[OUTFALLS]")
		coordsDict = inp.createDictionary("[COORDINATES]")
		allNodesDict = su.merge_dicts(storagesDict, junctionsDict, outfallsDict)

		if subset:
			allNodesDict = { key:value for key,value in allNodesDict.items() if key in subset }

		if rpt:
			#create a dictionary holding data from an rpt file, if provided
			nodesDepthSummaryDict = rpt.createDictionary("Node Depth Summary")
			nodesFloodSummaryDict = rpt.createDictionary("Node Flooding Summary")
			#print "created node summary dicts from " + rpt.fName

		maxEl = 0.00 #used for tranform
		minEl = 999999 #used for transform
		#errCount = 0
		node_objects = {}
		for node in allNodesDict:

			if not allNodesDict[node] or not node in coordsDict:
				continue

			#if the conduit's upstream & downstream nodes are
			# found in the allNodesDict look up it's data
			#try:
			invert = 	float(allNodesDict[node][0])
			xy = [float(i) for i in coordsDict[node]]#convert to floats

			if bbox and (not su.pointIsInBox(bbox, xy)):
				#skip nodes who are not within a given boudning box.
				continue

			n = Node(node, invert, xy)
			try:
			    n.maxDepth = float(allNodesDict[node][1])
			except ValueError:
			    pass #not a float, probably a 'FIXED' tag at that index on a outfall node


			if rpt and (node in nodesDepthSummaryDict):
				#if the up and downstream nodes are also found in the node depth summary dictionary,
				#grab the relevant data and store a little dictionary
				NDA = nodesDepthSummaryDict[node] #upstream node depth array

				n.maxHGL = float(NDA[3])
				#n.maxDepth = float(NDA[2])
				maxEl = max(n.maxHGL, maxEl) #increase the max elevation observed with the HGL

			if rpt and (node in nodesFloodSummaryDict):
				#if the up and downstream nodes are also found in the node flooding summary dictionary,
				#grab the relevant data and store a little dictionary
				NFA = nodesFloodSummaryDict[node] #upstream node flooding array
				n.flood_duration = float(NFA[0])
				#nodeDict.update({'floodDuration':floodDuration})

			#append data to the overall dictionary
			node_objects.update({node:n})

			maxEl = max(n.maxDepth, n.maxHGL, maxEl) #inverts plus the geom1 (pipe dimension)
			minEl = min(invert, minEl)

			#except:
				#errors occured when nodes aren't found in COORDINATES table
			#	print 'problem with ', node

		output = {
			'node_objects': node_objects,
			'maxEl':maxEl,
			'minEl':minEl,
			}

		#save for later use. save the bbox instance so we refresh this attributed
		#whenever the bbox is changed (so we don't leave elements out)
		self.organized_node_data = output
		self.bbox = bbox

		return output

	def node(self, node, conduit=None):

		#method for provide information about specific model elements
		#returns a node object given its ID
		if not self.organized_node_data:
			self.organized_node_data = self.organizeNodeData()

		n = self.organized_node_data['node_objects'][node]
		subcats_inp = self.inp.createDictionary("[SUBCATCHMENTS]")
		subcats_rpt = self.rpt.createDictionary('Subcatchment Runoff Summary')

		n.nodes_upstream = su.traceFromNode(self, node, mode='up')['nodes']
		n.subcats_direct = [k for k,v in subcats_inp.items() if v[1]==node]
		n.subcats_upstream = [k for k,v in subcats_inp.items() if v[1] in n.nodes_upstream]


		n.drainage_area_direct = sum([float(x) for x in [v[2] for k,v in subcats_inp.items() if k in n.subcats_direct]])
		n.drainage_area_upstream = sum([float(x) for x in [v[2] for k,v in subcats_inp.items() if k in n.subcats_upstream]])

		n.runoff_upstream_mg = sum([float(x) for x in [v[5] for k,v
								in subcats_rpt.items() if k in n.subcats_upstream]])
		n.runoff_upstream_cf = n.runoff_upstream_mg*1000000/7.48
		return n

	def list_objects(self,element_type='node', bbox=None, subset=None):
		"""
		return a dictionary of element obects of the type specified, with
		keys equal to the element ID.
		"""
		#organize the data -> dictionary of objects
		if element_type == 'node':
			data = self.organizeNodeData(bbox, subset=subset)
			dicts = data['node_objects']
		elif element_type == 'conduit':
			data = self.organizeConduitData(bbox, subset=subset)
			dicts = data['conduit_objects']
		elif element_type =='parcel':
			data = parcels.parcel_flood_duration(self, 'PWD_PARCELS_SHEDS', threshold=0)
			dicts =data['parcels']
		else:
			return "incorrect data type specified"

		return dicts

	def export_geojson(self, bbox=None, crs=None, filename=None):
		"""
		export model as a geojson object
		"""
		import geojson
		from geojson import Point, LineString, Feature, GeometryCollection

		if crs is None:
			#default coordinate system (http://spatialreference.org/ref/epsg/2272/):
			#NAD_1983_StatePlane_Pennsylvania_South_FIPS_3702_Feet
			crs =  {"type": "name","properties": {"name": "EPSG:2272"}}

		geometries = [] #array of features
		#collect the nodes
		for k,v in self.list_objects('node', bbox).items():
			props = {'flood_duration':v.flood_duration, 'id':v.id}
			geometry = Point(v.coordinates, properties=props)
			geometries.append(geometry)

		#collect the links
		for k,v in self.list_objects('conduit', bbox).items():
			props = {'MaxQPercent':v.maxQpercent,
			        'id':v.id,
			        'lifecycle':v.lifecycle,
			        'geom1':v.geom1,
					'geom2':v.geom2,}
			geometry = LineString(v.coordinates, properties=props)
			geometries.append(geometry)

		# points = []
		# for k, v in node_dicts.items():
		# 	xy = v.coordinates
		# 	points.append((xy[0], xy[1]))
		if filename is None:
			return GeometryCollection(geometries, crs=crs)
		else:
			with open(filename, 'wb') as f:
				f.write(geojson.dumps(GeometryCollection(geometries, crs=crs)))
				#f.write(geojson.dumps(FeatureCollection(nodefeatures, crs=crs)))


	def export_to_shapefile(self, element_type='node', filename=None, bbox=None, subset=None):
		"""
		export the model data into a shapefile. element_type dictates which type
		of data will be included.
		"""
		import shapefile
		import shutil

		#organize the data -> dictionary of objects
		dicts = self.list_objects(element_type, bbox, subset=subset)

		#grab first object in list to structure the dataframe
		ref_object = dicts[dicts.keys()[0]]
		fields = ref_object.__slots__
		data = [[getattr(v,j) for j in fields] for k,v in dicts.items()]

		#create a shp file writer object of geom type 'point'
		if element_type == 'node':
			w = shapefile.Writer(shapefile.POINT)
		elif element_type == 'conduit':
			w = shapefile.Writer(shapefile.POLYLINE)

		#use the helper mode to ensure the # of records equals the # of shapes
		#(shapefile are made up of shapes and records, and need both to be valid)
		w.autoBalance = 1

		#add the fields
		for fieldname in fields:
			w.field(fieldname, "C")

		#add the data
		for k, v in dicts.items():

			#add the coordinates
			xy = v.coordinates

			if element_type == 'node':
				w.point(xy[0], xy[1])
			elif element_type == 'conduit':
				w.line(parts = [xy])


			#add the record, by accessing each object attribute's data
			w.record(*[getattr(v,j) for j in fields])
			#print [getattr(v,j) for j in fields]
			#print '{}:  {} - {}'.format(v.id, xy[0], xy[1])

		w.save(filename)

		#save the projection data
		currentdir = os.path.dirname(__file__)
		prj_filepath = os.path.splitext(filename)[0] + '.prj'
		shutil.copy(os.path.join(currentdir, 'defs/default.prj'), prj_filepath)

	def export_to_csv(self,  element_type='node', filename=None, bbox=None, openfile=True):

		"""
		exports the organized SWMM data into a csv table
		"""

		#organize the data -> dictionary of objects
		dicts = self.list_objects(element_type, bbox)

		#grab first object in list to structure the dataframe
		ref_object = dicts[dicts.keys()[0]]
		data = [[getattr(v,j) for j in ref_object.__slots__] for k,v in dicts.items()]
		df = pd.DataFrame(data, columns = ref_object.__slots__)

		#save to file csv
		if not filename: filename = self.inp.name + "_" + element_type + 's_'
		csvfilepath = os.path.join(self.inp.dir, filename) + '.csv'
		df.to_csv(csvfilepath)

		if openfile:
			os.startfile(csvfilepath)

class SWMMIOFile(object):

	defaultSection = "Link Flow Summary"

	def __init__(self, filePath):

		#file name and path variables
		self.filePath = filePath
		self.fName = os.path.basename(filePath)
		self.name = os.path.splitext(self.fName)[0]
		self.dir = os.path.dirname(filePath)
		self.fileSize = os.path.getsize(filePath)


	def findByteRangeOfSection(self, startStr):

		#returns the start and end "byte" location of substrings in a text file

		with open(self.filePath) as f:
			start = None
			end = None
			l = 0 #line bytes index
			for line in f:

				#if start and len(line) <= 3 and (l - start) > 100:
				if start and line.strip() == "" and (l - start) > 100:
					#LOGIC ^ if start exists (was found) and the current line length is 3 or
					#less (length of /n ) and we're more than 100 bytes from the start location
					#then we are at the first "blank" line after our start section (aka the end of the section)
					end = l
					break

				if (startStr in line) and (not start):
					start = l

				l += len(line) + len("\n") #increment length (bytes?) of current position

		return [start, end]

	def exportCSV(self, sectionTitle = defaultSection):
		#create CSV of section of choice in inp file

		outFilePath = self.dir + "\\" + self.name + "_" + sectionTitle.replace(" ", "") + ".csv"
		outFile = open(outFilePath, 'w') #create outfile
		preppedTempFilePath = txt.extract_section_from_file(self.filePath, sectionTitle)
		if not preppedTempFilePath:
			return None #if nothing was found, do nothing

		with open(preppedTempFilePath) as rptFile:
			for line in rptFile:

				#if the line is blankish, break (chop off any junk at the end)
				if len(line) <=3: break
				if ";" in line: continue #don't include any comments
				#print line if there is stuff on it, CSV style
				line = ','.join(re.findall('\"[^\"]*\"|\S+', line))
				outFile.write("%s\n" % line)

		outFile.close()
		os.remove(preppedTempFilePath)

		return outFilePath

	def createDictionary (self, sectionTitle = defaultSection):

		"""
		Help info about this method.
		"""

		#preppedTempFilePath = self.readSectionAndCleanHeaders(sectionTitle) #pull relevant section and clean headers
		preppedTempFilePath = txt.extract_section_from_file(self.filePath, sectionTitle)
		if not preppedTempFilePath:
			return None #if nothing was found, do nothing

		passedHeaders = False

		with open(preppedTempFilePath) as file:
			the_dict = {}
			for line in file:

				if len(line) <=3 and not ";" in line: break
				if not passedHeaders:
					passedHeaders = True
					continue

				#check if line is commented out (having a semicolon before anything else) and skip accordingly
				if ";" == line.replace(" ", "")[0]:
					continue #omit this entire line

				line = line.split(";")[0] #don't look at anything to right of a semicolon (aka a comment)

				line = ' '.join(re.findall('\"[^\"]*\"|\S+', line))
				rowdata = line.replace("\n", "").split(" ")
				the_dict[str(rowdata[0])] = rowdata[1:] #create dictionary row with key and array of remaing stuff on line as the value

		os.remove(preppedTempFilePath)

		return the_dict




	def printSectionOfFile(self, lookUpStr=None, startByte=0, printLength = 500):

		l = startByte #line byte location
		with open(self.filePath) as f:

			f.seek(startByte) #jump to section at byte
			#return f.read(printLength)

			if lookUpStr:
				for line in f:
					#l += len(line) + len("\n")
					if lookUpStr in line:
						f.seek(l)
						break
					l += len(line) + len("\n")

			print f.read(printLength)

class rpt(SWMMIOFile):

	#creates an accessible SWMM .rpt object, inherits from SWMMIO object
	defaultImageDir = r"P:\Tools\Pipe Capacity Graphics\Scripts\image"
	def __init__(self, filePath):

		SWMMIOFile.__init__(self, filePath) #run the superclass init

		with open (filePath) as f:
			for line in f:
				if "Starting Date" in line:
					simulationStart = line.split(".. ")[1].replace("\n", "")
				if "Ending Date" in line:
					simulationEnd = line.split(".. ")[1].replace("\n", "")
				if "Report Time Step ........." in line:
					timeStepMin = int(line.split(":")[1].replace("\n", ""))
					break

		self.simulationStart = simulationStart
		self.simulationEnd = simulationEnd
		self.timeStepMin = timeStepMin

		#grab the date of analysis
		with open (filePath) as f:
			f.seek(self.fileSize - 500) #jump to 500 bytes before the end of file
			for line in f:
				if "Analysis begun on" in line:
					date = line.split("Analysis begun on:  ")[1].replace("\n", "")

		self.dateOfAnalysis = date

		#assign the header list
		#self.headerList = swmm_headers.rptHeaderList
		self.byteLocDict = None #populated if necessary elsewhere (LEGACY, can prob remove)
		self.elementByteLocations = {"Link Results":{}, "Node Results":{}} #populated if necessary elsewhere

	def createByteLocDict (self, sectionTitle = "Link Results"):

		#method creates a dictionary with Key = to SWMM element ID and
		#Value as the starting byte location of its time series in the rpt file
		#for rapidly accessing large rpt files

		#create set of other headers that are not the desired one, use to find end of section
		possibleNextSections = set(['Link Results', 'Node Results', 'Subcatchment Results']) - set([sectionTitle])

		print possibleNextSections

		startByte = self.findByteRangeOfSection(sectionTitle)[0] #+ len('\n  ************') #move past the first asterisks

		id_byteDict = {}
		with open(self.filePath) as f:

			f.seek(startByte) #jump to general area of file if we know it
			l = startByte
			for line in f:

				#if "<<<" in line and ">>>" in line:
				if "<<<" and ">>>" in line:	#cr
					#found the begining of a link's section
					lineCleaned = ' '.join(re.findall('\"[^\"]*\"|\S+', line))
					rowdata = lineCleaned.replace("\n", "").split(" ")

					#add to the dict
					id_byteDict.update({rowdata[2]:l})

				if any(header in line for header in possibleNextSections):
					#checks if line includes any of the other headers,
					#if so, we found next section, stop building dict
					break

				l += len(line) + len("\n") #increment length (bytes) of current position

		self.byteLocDict = id_byteDict
		self.elementByteLocations.update({sectionTitle:id_byteDict})
		return id_byteDict

	def returnDataAtDTime(self, id, dtime, sectionTitle="Link Results", startByte=0):

		#this is a slow ass function, when the file is big - can we improve this?
		byteLocDict = self.elementByteLocations[sectionTitle]
		if byteLocDict:
			startByte = byteLocDict[id]

		elif startByte == 0:
			startByte = self.findByteRangeOfSection(sectionTitle)[0]
			print 'startByte ' + str(startByte)

		with open(self.filePath) as f:

			f.seek(startByte) #jump to general area of file if we know it
			subsectionFound = False

			for line in f:
				if id in line: subsectionFound = True

				if subsectionFound and dtime in line:
					line = ' '.join(re.findall('\"[^\"]*\"|\S+', line))
					rowdata = line.replace("\n", "").split(" ")
					return rowdata

class inp(SWMMIOFile):

	#creates an accessible SWMM .inp object
	#make sure INP has been saved in the GUI before using this

	def __init__(self, filePath):

		SWMMIOFile.__init__(self, filePath) #run the superclass init

		#assign the header list
		#self.headerList = swmm_headers.inpHeaderList

	def replace_section(self, newdata, sectionheader='[FILES]'):

		"""
		replace a given section of the INP file with the contents of a passed
		dataframe object. Function creates a temporary copy of the original,
		rebuilds the entire inp, section by section, saves the temp file, then
		renames the temp file to the original file name.
		"""
		pass





class Node(object):

	#object representing a swmm node object

	__slots__ = ('id', 'invert', 'coordinates', 'nodes_upstream', 'subcats_direct', 'subcats_upstream',
				'drainage_area_direct', 'drainage_area_upstream',
				'runoff_upstream_cf', 'runoff_upstream_mg',
				'flood_duration', 'maxDepth',
				'maxHGL', 'draw_coordinates', 'lifecycle', 'delta_type', 'is_delta')

	def __init__(self, id, invert=None, coordinates=[]):

		#assign the header list
		self.id = id
		self.invert = invert
		self.coordinates = coordinates
		self.nodes_upstream = None
		self.subcats_direct = [] #subcatchment draining into this node, if any
		self.subcats_upstream = [] #all subcats upstream
		self.drainage_area_direct = None #only populated when needed
		self.drainage_area_upstream = None #only populated when needed
		self.runoff_upstream_cf = None #only populated when needed
		self.runoff_upstream_mg = None #only populated when needed
		self.flood_duration = 0
		self.maxDepth = 0
		self.maxHGL = 0
		self.draw_coordinates = None
		self.lifecycle = 'existing'
		self.is_delta = None
		self.delta_type = None #whether this object's data represents a change between two parent models

class Link(object):

	#object representing a swmm Link (conduit) object

	__slots__ = ('id', 'coordinates', 'maxflow', 'maxQpercent', 'upNodeID', 'downNodeID',
				'maxHGLDownstream', 'maxHGLUpstream', 'geom1', 'geom2', 'length','inletoffset', 'outletoffset',
				'draw_coordinates', 'lifecycle', 'delta_type', 'is_delta')


	def __init__(self, id, coordinates=[], geom1=None, geom2=None, inletoffset=0, outletoffset=0):

		#assign the header list
		self.id = id
		self.coordinates = coordinates
		self.maxflow = 0
		self.maxQpercent = 0
		self.maxHGLDownstream = 0
		self.maxHGLUpstream = 0
		self.inletoffset = inletoffset
		self.outletoffset = outletoffset
		self.upNodeID = None
		self.downNodeID = None
		self.geom1 = geom1 #faster to use zero as default?
		self.geom2 = geom2 #faster to use zero as default?
		self.length = None
		self.draw_coordinates = None
		self.lifecycle = 'existing'
		self.is_delta = None
		self.delta_type = None #whether this object's data represents a change between two parent models






#end



#!/usr/bin/env python
#coding:utf-8

#Utilities and such for SWMMIO processing
import math
from PIL import Image, ImageDraw, ImageFont, ImageOps
from time import strftime
import os
import numpy as np
# import matplotlib.path as mplPath
# from matplotlib.transforms import BboxBase
import pickle
import json
#import arcpy
from swmmio.graphics import draw_utils as du

#contants
sPhilaBox = 	((2683629, 220000), (2700700, 231000))
sPhilaSq =		((2683629, 220000), (2700700, 237071))
sPhilaSm1 = 	((2689018, 224343), (2691881, 226266))
sPhilaSm2 = 	((2685990, 219185), (2692678, 223831))
sPhilaSm3 = 	((2688842, 220590), (2689957, 221240))
sPhilaSm4 = 	((2689615, 219776), (2691277, 220738))
sPhilaSm5 = 	((2690303, 220581), (2690636, 220772))
sm6 = 			((2692788, 225853), (2693684, 226477))
chris = 		((2688798, 221573), (2702834, 230620))
nolibbox= 		((2685646, 238860),	(2713597, 258218))
mckean = 		((2691080, 226162),	(2692236, 226938))
d68d70 = 		((2691647, 221073),	(2702592, 227171))
d70 = 			((2694096, 222741),	(2697575, 225059))
ritner_moyamen =((2693433, 223967),	(2694587, 224737))
morris_10th = 	((2693740, 227260),	(2694412, 227693))
morris_12th = 	((2693257, 227422),	(2693543, 227614))
study_area = 	((2682005, 219180), (2701713, 235555))
dickenson_7th = ((2695378, 227948), (2695723, 228179))
packer_18th = 	((2688448, 219932), (2691332, 221857))
moore_broad = 	((2689315, 225537), (2695020, 228592))
oregon_front =  ((2695959, 221033), (2701749, 224921))
mckean_2nd = 	((2696719, 224600), (2699010, 226150))

#COLOR DEFS
red = 		(250, 5, 5)
blue = 		(5, 5, 250)
shed_blue = (0,169,230)
white =		(250,250,240)
black = 	(0,3,18)
lightgrey = (235, 235, 225)
grey = 		(100,95,97)
park_green = (115, 178, 115)
green = 	(115, 220, 115)
water_grey = (130, 130, 130)
purple = (250, 0, 250)

#FONTS
fontFile = r"C:\Data\Code\Fonts\Raleway-Regular.ttf"

def getFeatureExtent(feature, where="SHEDNAME = 'D68-C1'", geodb=r'C:\Data\ArcGIS\GDBs\LocalData.gdb'):

	import arcpy
	features = os.path.join(geodb, feature)
	for row in arcpy.da.SearchCursor(features, ["SHAPE@"], where_clause=where):

		#extent = row[0].extent
		#xy1,  xy2 =(extent.XMin, extent.YMin), (extent.XMax, extent.YMax)
		#return xy1, xy2
		#return row
		for part in row[0]:
			xy1,  xy2 =(part[0].X, part[0].Y), (part[1].X, part[1].Y)

			print xy1, xy2
			#print part

#FUNCTIONS
def traceFromNode(model, startNode, mode='up', stopnode=None):

	#nodes = model.organizeNodeData(bbox)['nodeDictionaries']
	#links = model.organizeConduitData(bbox)['conduitDictionaries']
	inp = model.inp
	conduitsDict = inp.createDictionary("[CONDUITS]")
	storagesDict = inp.createDictionary("[STORAGE]")
	junctionsDict = inp.createDictionary("[JUNCTIONS]")
	outfallsDict = inp.createDictionary("[OUTFALLS]")
	allNodesDict = merge_dicts(storagesDict, junctionsDict, outfallsDict)

	tracedNodes = [startNode] #include the starting node
	tracedConduits = []

	#recursive function to trace upstream
	def trace (nodeID):
		#print "tracing from {}".format(nodeID)
		for conduit, data in conduitsDict.iteritems():

			conduitUpNodeID = conduitDnNodeID = None
			if len(data) >= 1:
				#not sure why i need to do this check, but it prevents an indexing error on some
				conduitUpNodeID = data[0]
				conduitDnNodeID = data[1]

			if mode=='down' and conduitUpNodeID == nodeID and conduit not in tracedConduits:
				#conduit not in traced conduits to prevent duplicates for some reason
				#grab its dnstream node ID
				tracedConduits.append(conduit)
				tracedNodes.append(conduitDnNodeID)
				if stopnode and conduitDnNodeID == stopnode:
					break
				trace(conduitDnNodeID)

			if mode == 'up' and conduitDnNodeID == nodeID and conduit not in tracedConduits:
				#conduit not in traced conduits to prevent duplicates for some reason
				#this conduit is upstream of current node
				#grab its upstream node ID
				tracedConduits.append(conduit)
				tracedNodes.append(conduitUpNodeID)
				if stopnode and conduitUpNodeID == stopnode:
					break
				trace(conduitUpNodeID)

	#kickoff the trace
	print "Starting trace {0} from {1}".format(mode, startNode)
	trace(startNode)
	print "Traced {0} nodes from {1}".format(len(tracedNodes), startNode)
	return {'nodes':tracedNodes, 'conduits':tracedConduits}

def length_of_conduits(conduitsubset):

	#return the total length of conduits given a dictionary of Link objects
	l = 0
	for id, conduit in conduitsubset.iteritems():
		l +=conduit.length

	return l

def length_of_new_conduit(model1, model2):
    newconduitIDs = scomp.get_all_unmatched_inp_elements(model1, model2)['[CONDUITS]']['added']
    all_conduits = model2.organizeConduitData()['conduit_objects']
    l = 0
    for condID in newconduitIDs:
        l += all_conduits[condID].length

    return l



def pointIsInBox (bbox, point):

		#pass a lower left (origin) and upper right tuple representing a box,
		#and a tuple point to be tested

		LB = bbox[0]
		RU = bbox[1]

		x = point[0]
		y = point[1]

		if x < LB[0] or x >  RU[0]:
			return False
		elif y < LB[1] or y > RU[1]:
			return False
		else:
			return True

def pipeLengthPlanView(upstreamXY, downstreamXY):

	#return the distance (units based on input) between two points
	x1 = float(upstreamXY[0])
	x2 = float(downstreamXY[0])
	y1 = float(upstreamXY[1])
	y2 = float(downstreamXY[1])

	return math.hypot(x2 - x1, y2 - y1)
def pipeProfileLengthTransform(planLength, upstreamEl, downstreamEl, xTrans, yTrans):

	#tranform a pipe legth in x and y direction via its components

	ycomponent = upstreamEl - downstreamEl
	newXComponent = planLength * xTrans
	newYComponent = ycomponent * yTrans

	return math.hypot(newXComponent, newYComponent)
def merge_dicts(*dict_args):
    '''
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    '''
    result = {}
    for dictionary in dict_args:
		if dictionary:
			result.update(dictionary)
    return result


def elementChange(elementData, parameter='maxflow'):

	#with a joined dictionary item, returns change from existing to proposed

	proposedVal = elementData['proposed'].get(parameter, 0.0)
	existingVal = elementData['existing'].get(parameter, 0.0)

	return proposedVal - existingVal

def subsetConduitsInBoundingBox(conduitsDict, boundingBox):

	newDict = {}
	for conduit, conduitData in conduitsDict.iteritems():
		if 'existing' and 'proposed' in conduitData:
			#then we have a "compare" dictionary, must drill down further, use proposed
			coordPair = conduitData['proposed']['coordinates']
		else:
			coordPair = conduitData['coordinates']

		if boundingBox and (not pointIsInBox(boundingBox, coordPair[0]) or not pointIsInBox(boundingBox, coordPair[1])):
			#ship conduits who are not within a given boudning box
			continue

		newDict.update({conduit:conduitData})

	return 	newDict

def subsetElements(model, type='node', key='floodDuration', min=0.083, max=99999, bbox=None, pair_only=False):
	#return a subset of a dictionary of swmm elements based on a value

	if type=='node':
		elems = model.organizeNodeData(bbox)['node_objects']

	elif type=='conduit':
		elems = model.organizeConduitData(bbox)['conduit_objects']
	else: return []

	if pair_only:
		#only return the element and the value being filtered on
		subset = {k:v.flood_duration for (k,v) in elems.items() if v.flood_duration >= min }
	else:
		subset = {k:v for (k,v) in elems.items() if v.flood_duration >= min }

	return subset

def shape2Pixels(feature, cols = ["OBJECTID", "SHAPE@"],  where="SHEDNAME = 'D68-C1'", shiftRatio=None, targetImgW=1024, bbox=None, gdb=r'C:\Data\ArcGIS\GDBs\LocalData.gdb'):

	#take data from a geodatabase and organize in a dictionary with coordinates and draw_coordinates
	import arcpy
	features = os.path.join(gdb, feature)
	geometryDicts = {}
	for row in arcpy.da.SearchCursor(features, cols, where_clause=where):

		try:
			#detect what shape type this is
			geomType = row[1].type
			jsonkey = 'rings' #default for polygons, for accessing polygon vert coords
			if geomType == 'polyline':
				jsonkey = 'paths'


			geometrySections = json.loads(row[1].JSON)[jsonkey] # an array of parts
			#geomArr = json.loads(row[1].JSON)[jsonkey][0] #(assumes poly has one ring)

			for i, section in enumerate(geometrySections):

				#check if part of geometry is within the bbox, skip if not
				if bbox and len ( [x for x in section if pointIsInBox(bbox, x)] ) == 0:
					continue #skip this section if none of it is within the bounding box

				id = str(row[0])
				if len(geometrySections) > 1:
					id = str(row[0]) + "." + str(i)

				geometryDict = {'coordinates':section, 'geomType':geomType}
				#geometryDicts.update({id:{'coordinates':geomArr}})

				#add any other optional cols, starting at 3rd col item
				for j, col in enumerate(cols[2:]):
					col_data = str(row[j+2])
					geometryDict.update({col:col_data})

				geometryDicts.update({id:geometryDict})

		except:
			"prob with ", row[0]

	#find mins and maxs
	maxX = maxY = 0.0 #determine current width prior to ratio transform
	minX = minY = 99999999.0
	if not bbox:
		#review all points of the polygon to find min of all points
		for geom, geomData in geometryDicts.iteritems():
			for seg in geomData['coordinates']:
				minX = min(minX, seg[0])
				minY = min(minY, seg[1])
				maxX = max(maxX, seg[0])
				maxY = max(maxY, seg[1])
		bbox = [(minX, minY),(maxX, maxY)]

	#can probably get rid of these, if polygons are always secondary to the model drawing stuff (in which these params should be defined)
	height = float(bbox[1][1]) - float(bbox[0][1])
	width = float(bbox[1][0]) - float(bbox[0][0])
	shiftRatio = float(targetImgW / width) # to scale down from coordinate to pixels

	for geom, geomData in geometryDicts.iteritems():
		drawCoords = []
		for coord in geomData['coordinates']:

			drawCoord = coordToDrawCoord(coord, bbox, shiftRatio)
			drawCoords.append(drawCoord)
		geomData.update({'draw_coordinates':drawCoords})

	imgSize = (width*shiftRatio, height*shiftRatio)
	imgSize = [int(math.ceil(x)) for x in imgSize] #convert to ints

	polyImgDict = {'geometryDicts':geometryDicts, 'imgSize':imgSize , 'boundingBox': bbox, 'shiftRatio':shiftRatio }

	return polyImgDict

def convertCoordinatesToPixels(element_objs, targetImgW=1024, bbox=None, shiftRatio=None):

	#adds a dictionary to each conduit or node dict called
	#'draw_coordinates' which is a two part tuple, xy1, xy2

	if not bbox:
		#start by determining the max and min coordinates in the whole model
		maxX = maxY = 0.0 #determine current width prior to ratio transform
		minX = minY = 99999999
		for id, element in element_objs.iteritems():

			coordPair = element.coordinates

			if type(coordPair[0]) is list:
				#loop for elements with multiple coordinates (lines)
				#each coordinate is a list
				for coord in coordPair:
					minX = min(minX, coord[0])
					minY = min(minY, coord[1])
					maxX = max(maxX, coord[0])
					maxY = max(maxY, coord[1])

			else:
				minX = min(minX, coordPair[0])
				minY = min(minY, coordPair[1])
				maxX = max(maxX, coordPair[0])
				maxY = max(maxY, coordPair[1])

		bbox = [(minX, minY),(maxX, maxY)]

	height = float(bbox[1][1]) - float(bbox[0][1])
	width = float(bbox[1][0]) - float(bbox[0][0])
	if not shiftRatio:
		shiftRatio = float(targetImgW / width) # to scale down from coordinate to pixels

	for id, element in element_objs.iteritems():

		coordPair = element.coordinates

		drawcoords = []
		if type(coordPair[0]) is list:
			#loop for elements with multiple coordinates (lines)
			#each coordinate is a list
			for coord in coordPair:
				xy = coordToDrawCoord(coord, bbox, shiftRatio)
				drawcoords.append(xy)
		else:
			#zeroth index is not list, therefore coordPair is a single coord
			drawcoords = coordToDrawCoord(coordPair, bbox, shiftRatio)

		element.draw_coordinates = drawcoords

	imgSize = (width*shiftRatio, height*shiftRatio)
	imgSize = [int(math.ceil(x)) for x in imgSize] #convert to ints
	modelSizeDict = {'imgSize':imgSize, 'boundingBox': bbox, 'shiftRatio':shiftRatio }
	return modelSizeDict

def coordToDrawCoord(coordinates, bbox=None, shiftRatio=None, shiftRatioY=None, width=None, height=None):

	#convert single coordinate pair into the drawing space (pixels rather than cartesian coords)
	#given a cartesian bbox and shiftRatio

	#transform coords by normalizing by the mins, apply a ratio to
	#produce the desired width pixel space
	minX = minY = 0
	if bbox:
		minX = float(bbox[0][0])
		minY = float(bbox[0][1])
		maxX = float(bbox[1][0])
		maxY = float(bbox[1][1])

		height = maxY - minY
		width = maxX - minX

	if not shiftRatioY:
		shiftRatioY = shiftRatio
	x =  (coordinates[0] - minX) * shiftRatio
	y =  (height - coordinates[1] + minY) * shiftRatioY #subtract from height because PIL origin is in top right

	return (x,y)

def circleBBox(coordinates, radius):
	#returns the bounding box of a circle given as centriod coordinate and radius
	x = coordinates[0] #this indexing is because other elements haev more than on coordinate (ulgy pls fix)
	y = coordinates[1]
	r = radius

	return (x-r, y-r, x+r, y+r)


def drawNode(node, draw, options, rpt=None, dTime=None, xplier=1):

	color = (210, 210, 230) #default color
	radius = 0 #aka don't show this node by default
	outlineColor = None
	xy = node.draw_coordinates
	threshold = options['threshold']
	type = options['type']
	if dTime:

		data = rpt.returnDataAtDTime(node.id, dTime, sectionTitle="Node Results")
		q = abs(float(data[2])) #absolute val because backflow
		floodingQ = float(data[3])
		HGL = float(data[5])

		if type=='flood':

			#node params
			if floodingQ > 1:
				radius = q/2
				color = red #greenRedGradient(HGLup, 0, 15) #color

	elif node.is_delta:
		#we're dealing with "compare" object
		#floodDurationChange = elementChange(node, parameter='floodDuration')

		if node.flood_duration > 0 :
			#Flood duration increases
			radius = node.flood_duration*20
			color = red

		#if nodeData['existing'].get('floodDuration', 0) == 0 and nodeData['proposed'].get('floodDuration', 0) > 0:
		if node.delta_type == 'new_flooding':
			#new flooding found
			radius = node.flood_duration*20
			color = purple #purple
			outlineColor = (90, 90, 90)

	else:
		floodDuration = node.flood_duration
		maxDepth = node.maxDepth
		maxHGL = node.maxHGL

		if type == 'flood':
			#if floodDuration > 0.08333:
			if floodDuration >= threshold:
				radius = floodDuration*3
				color = red

		if type == 'flood_color':
			radius = 3
			if floodDuration >= threshold:
				yellow = (221, 220, 0)
				color = du.col2RedGradient(floodDuration, 0, 1, startCol=yellow) #options['fill'](floodDuration, 0, 1+threshold)
			else:

				radius = 1

	radius *= xplier
	draw.ellipse(circleBBox(xy, radius), fill =color, outline=outlineColor)

def drawConduit(conduit, draw, options, rpt=None, dTime = None, xplier = 1, highlighted=None):

	#Method for drawing (in plan view) a single conduit, given its RPT results

	#default fill and size
	fill = (120, 120, 130)
	drawSize = 1
	should_draw = True #boolean that can prevent a draw based on params
	coordPair = conduit.draw_coordinates
	type = options['type']
	#general method for drawing one conduit
	if rpt:
		#if an RPT is supplied, collect the summary data
		maxQPercent = conduit.maxQpercent
		q =  conduit.maxflow

		if maxQPercent !=0:
			capacity = q / maxQPercent
		else:
			capacity = 1
		stress = q / capacity

		if dTime:
			#if a datetime is provided, grab the specif flow data at this time
			data = rpt.returnDataAtDTime(conduit.id, dTime) #this is slow
			q = abs(float(data[2])) #absolute val because backflow
			stress = q / capacity    #how taxed is the pipe

		remaining_capacity = capacity-q
		#=================================
		#draw the conduit type as specifed
		#=================================
		if type == "flow":

			fill = options['fill']
			drawSize = options['draw_size'](q, options['exp'])

		if type == "flow_stress":

			fill = options['fill'](q*100, 0, capacity*175)
			drawSize = options['draw_size'](q, options['exp']) #int(round(math.pow(q, 0.67)))

		elif type == "stress":

			if maxQPercent >= 1:
				fill = options['fill'](q*100, 0, capacity*300) #greenRedGradient(q*100, 0, capacity*300)
				drawSize = options['draw_size'](stress*options['xplier'], options['exp']) #int(round(math.pow(stress*4, 1)))

		elif type == "capacity_remaining":

			if remaining_capacity > 0:
				fill = (0, 100, 255)
				drawSize = int( round( math.pow(remaining_capacity, 0.8)))



			#drawSize = int( round(max(remaining_capacity, 1)*xplier) )

	elif conduit.is_delta:
		#we're dealing with "compare" dictionary
		qChange = 	conduit.maxflow #elementChange(conduitData, parameter='maxflow')
		maxQperc = 	conduit.maxQpercent #elementChange(conduitData, parameter='maxQpercent')

		#default draw behavoir = draw grey lines proportional to geom1
		drawSize = min(7, conduit.geom1*0.7)
		fill = du.mediumgrey
		# #FIRST DRAW NEW OR CHANGED CONDUITS IN A CLEAR WAY
		# if conduit.lifecycle == 'new':
		# 	fill = blue
		# if conduit.lifecycle == 'changed':
		# 	fill = green #blue
			#drawSize = min(10, conduit.geom1)

		#IF THE CONDUITS IS 'EXISTING', DISPLAY SYMBOLOGY ACCORDINGLY (how things changed, etc)
		if conduit.lifecycle == 'existing':

			if type == 'proposed_simple':
				#drawSize = 0 #don't draw, only print the proposed infrastructure
				#fill = red
				#should_draw = False
				pass
			if type == 'compare_flow':
				drawSize = options['draw_size'](abs(qChange*options['xplier']), options['exp'])
				if qChange > 0:
					fill = du.greyRedGradient(qChange+15, 0, 20)
					#drawSize = int(round(math.pow(qChange, 0.67)))

				if qChange <= 0:
					fill = du.greyGreenGradient(abs(qChange)+15, 0, 20)
					#drawSize = int(round(math.pow(abs(qChange), 0.67)))

			if type == 'compare_hgl':

				avgHGL = (conduit.maxHGLUpstream + conduit.maxHGLDownstream) / 2.0
				drawSize = options['draw_size'](abs(avgHGL*options['xplier']), options['exp'])
				if avgHGL > 0:
					fill = du.red #du.greyRedGradient(avgHGL+15, 0, 20)
					#drawSize = int(round(math.pow(avgHGL, 1)))

				if avgHGL <= 0:
					fill = du.green_bright #du.greyGreenGradient(abs(avgHGL)+15, 0, 20)
					#drawSize = int(round(math.pow(avgHGL, 1)))


	#if highlighted list is provided, overide any symbology for the highlighted conduits
	if highlighted and conduit.id in highlighted:
		fill = blue
		drawSize = 3
	#FIRST DRAW NEW OR CHANGED CONDUITS IN A CLEAR WAY
	if conduit.lifecycle == 'new':
		fill = blue
	if conduit.lifecycle == 'changed':
		fill = green #blue
	drawSize = int(drawSize*xplier)


	if should_draw:
		#draw that thing
		draw.line(coordPair, fill = fill, width = drawSize)
		if pipeLengthPlanView(coordPair[0], coordPair[1]) > drawSize*0.75:
			#if length is long enough, add circles on the ends to smooth em out
			#this check avoids circles being drawn for tiny pipe segs
			draw.ellipse(du.circleBBox(coordPair[0], drawSize*0.5), fill =fill)
			draw.ellipse(du.circleBBox(coordPair[1], drawSize*0.5), fill =fill)

def angleBetweenPoint(xy1, xy2):
	dx, dy = (xy2[0] - xy1[0]), (xy2[1] - xy1[1])

	angle = (math.atan(float(dx)/float(dy)) * 180/math.pi )
	if angle < 0:
		angle = 270 - angle
	else:
		angle = 90 - angle
	#angle in radians
	return angle

def midPoint(xy1, xy2):

	dx, dy = (xy2[0] + xy1[0]), (xy2[1] + xy1[1])
	midpt = ( int(dx/2), int(dy/2.0) )

	#angle in radians
	return midpt

def annotateLine (img, dataDict, fontScale=1, annoKey='name', labeled = None):

	txt = dataDict[annoKey]

	if not txt in labeled:
		#do not repeat labels
		font = ImageFont.truetype(fontFile, int(25 * fontScale))
		imgTxt = Image.new('L', font.getsize(txt))
		drawTxt = ImageDraw.Draw(imgTxt)
		drawTxt.text((0,0), txt, font=font, fill=(10,10,12))

		coords = dataDict['coordinates']
		drawCoord = dataDict['draw_coordinates']
		angle = angleBetweenPoint(coords[0], coords[1])
		texRot = imgTxt.rotate(angle, expand=1)
		#canvas.paste( ImageOps.colorize(texRot, (0,0,0), (255,255,84)), (242,60),  texRot)

		midpoint = midPoint(drawCoord[0], drawCoord[1])
		#img.paste(texRot , midpoint,  texRot)
		img.paste(ImageOps.colorize(texRot, (0,0,0), (10,10,12)), midpoint,  texRot)
		labeled.append(txt) #keep tracj of whats been labeled



#def drawAnnotation (canvas, inp, rpt=None, imgWidth=1024, title=None, currentTstr = None, description = None,
#					objects = None, symbologyType=None, fill=(50,50,50), xplier=None):
def annotateMap (canvas, model, model2=None, currentTstr = None, options=None, results={}):

	#unpack the options
	nodeSymb = 		options['nodeSymb']
	conduitSymb = 	options['conduitSymb']
	basemap = 		options['basemap']
	parcelSymb = 	options['parcelSymb']
	traceUpNodes =	options['traceUpNodes']
	traceDnNodes =	options['traceDnNodes']

	modelSize = (canvas.im.getbbox()[2], canvas.im.getbbox()[3])

	#define main fonts
	fScale = 1 * modelSize[0] / 2048
	titleFont = ImageFont.truetype(fontFile, int(40 * fScale))
	font = ImageFont.truetype(fontFile, int(20 * fScale))

	#Buid the title and files list (handle 1 or two input models)
	#this is hideous, or elegant?
	files = title = results_string = symbology_string = annotationTxt = ""
	files = '\n'.join([m.rpt.filePath for m in filter(None, [model, model2])])
	title = ' to '.join([m.inp.name for m in filter(None, [model, model2])])
	symbology_string = ', '.join([s['title'] for s in filter(None, [nodeSymb, conduitSymb, parcelSymb])])
	title += "\n" + symbology_string

	#params_string = ''.join([" > " + str(round(s['threshold']*600)/10) + "min " for s in filter(None, [parcelSymb])])
	#build the title


	#collect results
	for result, value in results.iteritems():
		results_string += '\n' + result + ": " + str(value)

	#compile the annotation text
	if results:
		annotationTxt = results_string + "\n"
	annotationTxt += files


	annoHeight = canvas.textsize(annotationTxt, font)[1]

	canvas.text((10, 15), title, fill=black, font=titleFont)
	canvas.text((10, modelSize[1] - annoHeight - 10), annotationTxt, fill=black, font=font)


	if currentTstr:
		#timestamp in lower right corner
		annoHeight = canvas.textsize(currentTstr, font)[1]
		annoWidth = canvas.textsize(currentTstr, font)[0]
		canvas.text((modelSize[0] - annoWidth - 10, modelSize[1] - annoHeight - 10), currentTstr, fill=black, font=font)

	#add postprocessing timestamp
	timestamp = strftime("%b-%d-%Y %H:%M:%S")
	annoHeight = canvas.textsize(timestamp, font)[1]
	annoWidth = canvas.textsize(timestamp, font)[0]
	canvas.text((modelSize[0] - annoWidth - 10, annoHeight - 5), timestamp, fill=du.grey, font=font)




#end
